; 65c816 opcodes 0x00-0x0F:
;    00: BRK s       2,8
;    01: ORA (d,x)   2,6
;    02: COP s       2,8
;    03: ORA d,s     2,4
;    04: TSB d       2,5
;    05: ORA d       2,3
;    06: ASL d       2,5
;    07: ORA [d]     2,6
;    08: PHP s       1,3
;    09: ORA #       2,2
;    0A: ASL a       1,2
;    0B: PHD s       1,4
;    0C: TSB a       3,6
;    0D: ORA a       3,4
;    0E: ASL a       3,6
;    0F: ORA al      4,5

;%1 = vector, %2 = offset
%macro NativeInterrupt 2
 mov al,B_PB
 E0_PUSH_B
 GET_PC eax
 add eax,byte 2
 E0_PUSH_W
 E0_SETUPFLAGS          ; Put flags into true 65c816 format
 E0_PUSH_B

 mov R_NativePC,%1      ; PC = Native mode vector
 mov eax,%2
 mov B_PB,byte 0        ; Setup bank
 mov B_PBOffset,eax     ; Offset of bank 0 ROM
 SET_FLAG SNES_FLAG_I   ; Disable IRQs
 add R_NativePC,eax
 mov eax,[_Read_BankMapping]
 mov [CycleTable],byte 4    ; SlowROM bank
 CLR_FLAG SNES_FLAG_D       ; Disable decimal mode
 mov B_PB_Read,eax          ; Bank 0 read through memory mapper
 jmp near CPU_RETURN
%endmacro

;%1 = vector, %2 = offset
%macro Emu6502Interrupt 2
 GET_PC eax
 add eax,byte 2
 E1_PUSH_W
 E1_SETUPFLAGS          ; Put flags into true 65c816 format
 E1_PUSH_B

 mov R_NativePC,%1      ; PC = Native mode vector
 mov eax,%2
 mov B_PB,byte 0        ; Setup bank
 mov B_PBOffset,eax     ; Offset of bank 0 ROM
 SET_FLAG SNES_FLAG_I   ; Disable IRQs
 add R_NativePC,eax
 mov eax,[_Read_BankMapping]
 mov [CycleTable],byte 4    ; SlowROM bank
 CLR_FLAG SNES_FLAG_D       ; Disable decimal mode
 mov B_PB_Read,eax          ; Bank 0 read through memory mapper
 jmp near CPU_RETURN
%endmacro

; 00

ALIGNC
E0_BRK:
 NativeInterrupt B_BRK_Nvector,B_BRK_Noffset

ALIGNC
E1_BRK:
 Emu6502Interrupt B_IRQ_Evector,B_IRQ_Eoffset

; 01
EM_ORA Od_xO,Direct_Indexed_Indirect
E0_ORA Od_xO,Direct_Indexed_Indirect

; 02

ALIGNC
E0_COP:
 NativeInterrupt B_COP_Nvector,B_COP_Noffset

ALIGNC
E1_COP:
 Emu6502Interrupt B_COP_Evector,B_COP_Eoffset

; 03
EM_ORA d_s,Stack_Relative
E0_ORA d_s,Stack_Relative

; 04
EM_TSB d,Direct
E0_TSB d,Direct

; 05
EM_ORA d,Direct
E0_ORA d,Direct

; 06
EM_ASL d,Direct
E0_ASL d,Direct

; 07
EM_ORA IdI,Direct_Indirect_Long
E0_ORA IdI,Direct_Indirect_Long

; 08

ALIGNC
E0_PHP:
 inc R_NativePC
 E0_SETUPFLAGS
 E0_PUSH_B
 jmp near CPU_RETURN
ALIGNC
E1_PHP:
 inc R_NativePC
 E1_SETUPFLAGS
 E1_PUSH_B
 jmp near CPU_RETURN

; 09
EM_ORA i,Immediate
E0_ORA i,Immediate

; 0A

ALIGNC
EM_SLA:
 inc R_NativePC
 shl byte B_A,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN
ALIGNC
E0_SLA:
 inc R_NativePC
 shl word B_A,byte 1
 STORE_FLAGS SNES_FLAG_NZC
 jmp near CPU_RETURN

; 0B
ALIGNC
E0_PHD:
 inc R_NativePC
 mov eax,B_D
 E0_PUSH_W
 jmp near CPU_RETURN
ALIGNC
E1_PHD:
 inc R_NativePC
 mov eax,B_D
 E1_PUSH_W
 jmp near CPU_RETURN

; 0C
EM_TSB a,Absolute
E0_TSB a,Absolute

; 0D
EM_ORA a,Absolute
E0_ORA a,Absolute

; 0E
EM_ASL a,Absolute
E0_ASL a,Absolute

; 0F
EM_ORA al,Absolute_Long
E0_ORA al,Absolute_Long
