; SNEeSe 65c816 CPU emulation core
; Originally written by Savoury SnaX (Not quite sure if I like AT&T)
; Maintained/rewritten by Charles Bilyue'
;
; Compile under DJGPP (GNU-AS)
;  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!
;
; This file contains:
;  65c816 addressing mode macros
;
; Note: As of 0.372 (11-99), be very careful when modifying this code,
; and the CPU loops! The CPU loops set eax/ebx to 0, except al/bl, and
; esi to the address of the data at PB:PC (opcode + operands),
; and the new address mode emulation expects this!
;
; Final address in ebx, or eax when address is to be stored in PC!

; Immediate (read-only)
;  Read 8-bit and 16-bit
;    ORA-09  AND-29  EOR-49  ADC-69  BIT-89  LDY-A0  LDX-A2  LDA-A9
;    CPY-C0  CMP-C9  CPX-E0  SBC-E9
;  Read 8-bit
;    REP-C2  SEP-E2
;
%macro ADDR_Immediate 0
%endmacro

%macro READ8_Immediate 0
 mov al,[1+R_NativePC]
 add R_NativePC,byte 2
%endmacro

%macro READ16_Immediate 0
 mov al,[1+R_NativePC]
 mov ah,[2+R_NativePC]
 add R_NativePC,byte 3
%endmacro

; Absolute
;  Address = DB:a16
;  Read 8-bit and 16-bit
;    ORA-0D  BIT-2C  AND-2D  EOR-4D  ADC-6D  LDY-AC  LDA-AD  LDX-AE
;    CPY-CC  CMP-CD  CPX-EC  SBC-ED
;  Write 8-bit and 16-bit
;    STZ-9C  STY-8C  STA-8D  STX-8E
;  RMW 8-bit and 16-bit
;    TSB-0C  ASL-0E  TRB-1C  ROL-2E  LSR-4E  ROR-6E  DEC-CE  INC-EE
;
%macro ADDR_Absolute 0
 mov ebx,B_DB_Shifted
 mov bl,[1+R_NativePC]
 mov bh,[2+R_NativePC]
 add R_NativePC,byte 3
%endmacro

%macro READ8_Absolute 0
 ADDR_Absolute
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Absolute 0-1 0
 ADDR_Absolute
 GET_WORD %1
%endmacro

; Absolute Long (corrupts eax)
;  Address = a24
;  Read 8-bit and 16-bit
;    ORA-0F  AND-2F  EOR-4F  ADC-6F  LDA-AF  CMP-CF  SBC-EF
;  Write 8-bit and 16-bit
;    STA-8F
;
%macro ADDR_Absolute_Long 0
 mov bl,[3+R_NativePC]
 shl ebx,16
 mov bl,[1+R_NativePC]
 mov bh,[2+R_NativePC]
 add R_NativePC,byte 4
%endmacro

%macro READ8_Absolute_Long 0
 ADDR_Absolute_Long
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Absolute_Long 0-1 0
 ADDR_Absolute_Long
 GET_WORD %1
%endmacro

; Direct (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-05  BIT-24  AND-25  EOR-45  ADC-65  LDY-A4  LDA-A5  LDX-A6
;    CPY-C4  CMP-C5  CPX-E4  SBC-E5
;  Write 8-bit and 16-bit
;    STZ-64  STY-84  STA-85  STX-86
;  RMW 8-bit and 16-bit
;    TSB-04  ASL-06  TRB-14  ROL-26  LSR-46  ROR-66  DEC-C6  INC-E6
;
%macro ADDR_Direct 0
 mov bl,[1+R_NativePC]
 add R_NativePC,byte 2
 add bx,B_D
%endmacro

%macro READ8_Direct 0
 ADDR_Direct
 GET_BYTE_00
%endmacro

;%1 = Reverse
%macro READ16_Direct 0-1 0
 ADDR_Direct
 GET_WORD_00 %1
%endmacro

; Direct Indirect Indexed (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-11  AND-31  EOR-51  ADC-71  LDA-B1  CMP-D1  SBC-F1
;  Write 8-bit and 16-bit
;    STA-91
;
%macro ADDR_Direct_Indirect_Indexed 0
 READ16_Direct Reverse
 mov ebx,B_DB_Shifted
 mov bh,al
 mov bl,ah
 add ebx,B_Y
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Direct_Indirect_Indexed 0
 ADDR_Direct_Indirect_Indexed
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Direct_Indirect_Indexed 0-1 0
 ADDR_Direct_Indirect_Indexed
 GET_WORD %1
%endmacro

; Direct Indirect Indexed Long (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-17  AND-37  EOR-57  ADC-77  LDA-B7  CMP-D7  SBC-F7
;  Write 8-bit and 16-bit
;    STA-97
;
%macro ADDR_Direct_Indirect_Indexed_Long 0
 READ16_Direct Reverse
 bswap eax
 inc bx
 GET_BYTE_00
 ror eax,16
 mov ebx,B_Y
 add ebx,eax
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Direct_Indirect_Indexed_Long 0
 ADDR_Direct_Indirect_Indexed_Long
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Direct_Indirect_Indexed_Long 0-1 0
 ADDR_Direct_Indirect_Indexed_Long
 GET_WORD %1
%endmacro

; Direct index X (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-15  BIT-34  AND-35  EOR-55  ADC-75  LDY-B4  LDA-B5  CMP-D5
;    SBC-F5
;  Write 8-bit and 16-bit
;    STZ-74  STY-94  STA-95
;  RMW 8-bit and 16-bit
;    ASL-16  ROL-36  LSR-56  ROR-76  DEC-D6  INC-F6
;
%macro ADDR_Direct_Index_X 0
 mov bl,[1+R_NativePC]
 add R_NativePC,byte 2
 add bx,B_D
 add bx,B_X
%endmacro

%macro READ8_Direct_Index_X 0
 ADDR_Direct_Index_X
 GET_BYTE_00
%endmacro

;%1 = Reverse
%macro READ16_Direct_Index_X 0-1 0
 ADDR_Direct_Index_X
 GET_WORD_00 %1
%endmacro

; Direct index Y (corrupts eax)
;  Read 8-bit and 16-bit
;    LDX-B6
;  Write 8-bit and 16-bit
;    STX-96
;
%macro ADDR_Direct_Index_Y 0
 mov bl,[1+R_NativePC]
 add R_NativePC,byte 2
 add bx,B_D
 add bx,B_Y
%endmacro

%macro READ8_Direct_Index_Y 0
 ADDR_Direct_Index_Y
 GET_BYTE_00
%endmacro

;%1 = Reverse
%macro READ16_Direct_Index_Y 0-1 0
 ADDR_Direct_Index_Y
 GET_WORD_00 %1
%endmacro

; Direct Indexed Indirect (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-01  AND-21  EOR-41  ADC-61  LDA-A1  CMP-C1  SBC-E1
;  Write 8-bit and 16-bit
;    STA-81
;
%macro ADDR_Direct_Indexed_Indirect 0
 READ16_Direct_Index_X Reverse
 mov ebx,B_DB_Shifted
 mov bh,al
 mov bl,ah
%endmacro

%macro READ8_Direct_Indexed_Indirect 0
 ADDR_Direct_Indexed_Indirect
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Direct_Indexed_Indirect 0-1 0
 ADDR_Direct_Indexed_Indirect
 GET_WORD %1
%endmacro

; Absolute Index X (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-1D  BIT-3C  AND-3D  EOR-5D  ADC-7D  LDY-BC  LDA-BD  CMP-DD
;    SBC-FD
;  Write 8-bit and 16-bit
;    STZ-9E  STA-9D
;  RMW 8-bit and 16-bit
;    ASL-1E  ROL-3E  LSR-5E  ROR-7E  DEC-DE  INC-FE
;
%macro ADDR_Absolute_Index_X 0
 mov ebx,B_DB_Shifted
 mov bl,[1+R_NativePC]
 mov eax,B_X
 mov bh,[2+R_NativePC]
 add R_NativePC,byte 3
 add ebx,eax
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Absolute_Index_X 0
 ADDR_Absolute_Index_X
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Absolute_Index_X 0-1 0
 ADDR_Absolute_Index_X
 GET_WORD %1
%endmacro

; Absolute Index Y (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-19  AND-39  EOR-59  ADC-79  LDA-B9  LDX-BE  CMP-D9  SBC-F9
;  Write 8-bit and 16-bit
;    STA-99
;
%macro ADDR_Absolute_Index_Y 0
 mov ebx,B_DB_Shifted
 mov bl,[1+R_NativePC]
 mov eax,B_Y
 mov bh,[2+R_NativePC]
 add R_NativePC,byte 3
 add ebx,eax
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Absolute_Index_Y 0
 ADDR_Absolute_Index_Y
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Absolute_Index_Y 0-1 0
 ADDR_Absolute_Index_Y
 GET_WORD %1
%endmacro

; Absolute Long Index X (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-1F  AND-3F  EOR-5F  ADC-7F  LDA-BF  CMP-DF  SBC-FF
;  Write 8-bit and 16-bit
;    STA-9F
;
%macro ADDR_Absolute_Long_Index_X 0
 mov bl,[3+R_NativePC]
 shl ebx,16
 mov bl,[1+R_NativePC]
 mov eax,B_X
 mov bh,[2+R_NativePC]
 add ebx,eax
 add R_NativePC,byte 4
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Absolute_Long_Index_X 0
 ADDR_Absolute_Long_Index_X
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Absolute_Long_Index_X 0-1 0
 ADDR_Absolute_Long_Index_X
 GET_WORD %1
%endmacro

; Direct Indirect (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-12  AND-32  EOR-52  ADC-72  LDA-B2  CMP-D2  SBC-F2
;  Write 8-bit and 16-bit
;    STA-92
;  Push address (doesn't use these macros)
;    PEI-D4
;
%macro ADDR_Direct_Indirect 0
 READ16_Direct Reverse
 mov ebx,B_DB_Shifted
 mov bh,al
 mov bl,ah
%endmacro

%macro READ8_Direct_Indirect 0
 ADDR_Direct_Indirect
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Direct_Indirect 0-1 0
 ADDR_Direct_Indirect
 GET_WORD %1
%endmacro

; Direct Indirect Long (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-07  AND-27  EOR-47  ADC-67  LDA-A7  CMP-C7  SBC-E7
;  Write 8-bit and 16-bit
;    STA-87
;
%macro ADDR_Direct_Indirect_Long 0
 READ16_Direct Reverse
 bswap eax
 inc bx
 GET_BYTE_00
 ror eax,16
 mov ebx,eax
%endmacro

%macro READ8_Direct_Indirect_Long 0
 ADDR_Direct_Indirect_Long
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Direct_Indirect_Long 0-1 0
 ADDR_Direct_Indirect_Long
 GET_WORD %1
%endmacro

; Stack Relative (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-03  AND-23  EOR-43  ADC-63  LDA-A3  CMP-C3  SBC-E3
;  Write 8-bit and 16-bit
;    STA-83
;
%macro ADDR_Stack_Relative 0
 mov al,[1+R_NativePC]
 mov ebx,B_S    ; Add stack pointer to make address
 add R_NativePC,byte 2
 add bl,al
 adc bh,ah
%endmacro

%macro READ8_Stack_Relative 0
 ADDR_Stack_Relative
 GET_BYTE_00
%endmacro

;%1 = Reverse
%macro READ16_Stack_Relative 0-1 0
 ADDR_Stack_Relative
 GET_WORD_00 %1
%endmacro

; Stack Relative Indirect Indexed (corrupts eax)
;  Read 8-bit and 16-bit
;    ORA-13  AND-33  EOR-53  ADC-73  LDA-B3  CMP-D3  SBC-F3
;  Write 8-bit and 16-bit
;    STA-93
;
%macro ADDR_Stack_Relative_Indirect_Indexed 0
 ADDR_Stack_Relative
 mov eax,B_DB_Shifted
 GET_WORD
 mov ebx,B_Y
 add ebx,eax
 and ebx,0x00FFFFFF ; Ensure 24-bit address
%endmacro

%macro READ8_Stack_Relative_Indirect_Indexed 0
 ADDR_Stack_Relative_Indirect_Indexed
 GET_BYTE
%endmacro

;%1 = Reverse
%macro READ16_Stack_Relative_Indirect_Indexed 0-1 0
 ADDR_Stack_Relative_Indirect_Indexed
 GET_WORD %1
%endmacro

; Absolute Indexed Indirect (corrupts eax)
; Returns new PC in %ax
;  Address = (PB:X+a16)
;  Address set to PC
;    JMP-7C  JSR-FC
;
%macro ADDR_Absolute_Indexed_Indirect 0
 mov ebx,B_PB_Shifted
 mov bl,[1+R_NativePC]
 mov bh,[2+R_NativePC]
 add bx,B_X
 GET_BYTE_PB
 inc bx
 mov ah,al
 GET_BYTE_PB
 xchg ah,al
%endmacro

; Absolute (corrupts eax)
;  Address = a16
;  Address set to PC
;    JSR-20  JMP-4C
;
%macro ADDR_Absolute_JMP 0
 mov al,[1+R_NativePC]
 mov ah,[2+R_NativePC]
%endmacro
