#ifndef SNEeSe_DMA_i
#define SNEeSe_DMA_i

/*

 (H)DMA macros

 Thanks to zsKnight for all the information which helped get this
 working right! (I hope!)

*/

// Offsets from TableDMA#
#define DMAP 0
#define BBAD 1
#define DMA_Vid BBAD
#define NTRL 2
#define DMA_Inc 3
#define A1T 4
#define A1TL A1T
#define A1TH (A1T+1)
#define A1B (A1T+2)
#define DAS 8
#define DASL DAS
#define DASH (DAS+1)
#define DASB (DAS+2)
#define A2T 0x0C
#define A2L A2T
#define A2H (A2T+1)
#define A2B (A2T+2)
#define DMA_Wr0 0x10
#define DMA_Wr1 0x14
#define DMA_Wr2 0x18
#define DMA_Wr3 0x1C
#define DMA_Rd0 0x20
#define DMA_Rd1 0x24
#define DMA_Rd2 0x28
#define DMA_Rd3 0x2C
#define ABank_Rd 0x30
#define ABank_Wr 0x34
#define IBank_Rd 0x38
#define HDMA_Siz 0x3C
#define DMATableSize 0x40

.macro MAP_READ_DMA_LIST num
  .long   MAP_READ_DMAP\num
  .long   MAP_READ_BBAD\num
  .long   MAP_READ_A1TL\num
  .long   MAP_READ_A1TH\num
  .long   MAP_READ_A1B\num
  .long   MAP_READ_DASL\num
  .long   MAP_READ_DASH\num
  .long   MAP_READ_DASB\num
  .long   MAP_READ_A2L\num
  .long   MAP_READ_A2H\num
  .long   MAP_READ_NTRL\num
  DUPLICATE long,5,UNSUPPORTED_READ
.endm

.macro MAP_WRITE_DMA_LIST num
  .long   MAP_WRITE_DMAP\num
  .long   MAP_WRITE_BBAD\num
  .long   MAP_WRITE_A1TL\num
  .long   MAP_WRITE_A1TH\num
  .long   MAP_WRITE_A1B\num
  .long   MAP_WRITE_DASL\num
  .long   MAP_WRITE_DASH\num
  .long   MAP_WRITE_DASB\num
  .long   MAP_WRITE_A2L\num
  .long   MAP_WRITE_A2H\num
  .long   MAP_WRITE_NTRL\num
  DUPLICATE long,5,UNSUPPORTED_WRITE
.endm

.macro LOAD_DMA_TABLE num
    movl $TableDMA\num,%edx
.endm

.macro HDMAOPERATION num
  movb HDMAON,%al
  testb $(1<<\num),%al
  jz  8f

  movb $\num,In_DMA
  LOAD_DMA_TABLE \num
  call Do_HDMA_Channel  // CF clear if channel disabled

  jc  8f
  movb HDMAON,%al
  andb $~(1<<\num),%al
  movb %al,HDMAON       // Disable this channel
8:
.endm

.macro RELATCH_HDMA_CHANNEL num
    testb $(1<<\num),_HDMAEN
    jz  8f

    movb $0,NTRL_\num

/*
#ifdef DEBUG
#ifdef HDMATRACKER
    movl _LastHDMA,%ebx     //
    addl _HDMAAddress,%ebx  //
#endif
#endif
 */
    movl A1T_\num,%eax   // Src Address in ebx
    movl %eax,A2T_\num

/*
#ifdef DEBUG
#ifdef HDMATRACKER
    bswapl %eax             //
    movl %eax,4(%ebx)       //
    movb DASB_\num,%al      //
    movb %al,2(%ebx)        //
#endif
#endif

#ifdef DEBUG
#ifdef HDMATRACKER
    movb %al,(%ebx)         //
#endif
#endif

#ifdef DEBUG
#ifdef HDMATRACKER
    movb %dl,1(%ebx)        //
#endif
#endif

#ifdef DEBUG
#ifdef HDMATRACKER
    movl _LastHDMA,%ebx     //
    addl $8,%ebx            //
    andl $0x3FF,%ebx        //
    movl %ebx,_LastHDMA     //
#endif
#endif
 */
8:
.endm

.macro RELATCH_HDMA_IN_FRAME num
    movb HDMAON,%al
    notb %al
    andb (%esp),%al
    testb $(1<<\num),%al
    jz  8f

//  movb $0,NTRL_\num
8:
.endm

.macro RELATCH_HDMA
    movb _HDMAEN,%al
    movb %al,HDMAON
    testb %al,%al
    jz  9f
    RELATCH_HDMA_CHANNEL 0
    RELATCH_HDMA_CHANNEL 1
    RELATCH_HDMA_CHANNEL 2
    RELATCH_HDMA_CHANNEL 3
    RELATCH_HDMA_CHANNEL 4
    RELATCH_HDMA_CHANNEL 5
    RELATCH_HDMA_CHANNEL 6
    RELATCH_HDMA_CHANNEL 7
9:
.endm

#endif /* SNEeSe_DMA_i */
