/*

 EmuGUI.cc
  Contains emulator specific GUI code
  Some stuff from helper.c will be moved here!

*/

#define VAR_CYCLE_RATIO

//#include <fstream.h>
//#include <dpmi.h>
//#include <go32.h>
//#include <dos.h>
#include <stdio.h>
//#include <stdlib.h>
//#include <conio.h>
#include <unistd.h>

#include "wrapaleg.h"   // Wrapper header for Allegro
#include "guicore.h"

/* ------------------------- GUI STUFF ------------------------- */

#include "types.h"
#include "font.h"
#include "helper.h"
#include "CPU.h"
#include "SPC.h"
#include "sound.h"
#include "romload.h"

BITMAP *joypad=0;
extern PALETTE sneesepal;

char *GUI_init(){
 char *errormsg=GUI_core_init();

 if (errormsg) return errormsg;

 #ifdef DEBUG
  joypad = load_pcx("sneese.dat#joypad",sneesepal);
 #else
  joypad = load_pcx("#joypad",sneesepal);
 #endif

 return 0;
}

WINDOW *GUI_window=0;

extern BITMAP *sneese;

WINDOW *Main_window=0;
#define NUM_OPTIONS 8
char *Options[NUM_OPTIONS]={
 "Resume emulation",
 "Reset emulation",
 "Load ROM",
 "Configure",
 "ROM information",
 "HW status",
 "(H)DMA",
 "Exit SNEeSe :("};

WINDOW *Credits_window=0;

void UpdateGUI(int Selected){
 clear(GUI_Bitmap);
 Main_window->refresh();
 for(int a=0;a<NUM_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Main_window,default_font,Options[a],0,default_font->get_heightspace()*a,16);
  else PlotSelectedMenuItem(Main_window,default_font,Options[a],0,default_font->get_heightspace()*a,16);
 Credits_window->refresh();
 PlotString(Credits_window,default_font," Code by: Charles Bilyue'      ",0,0);
 PlotString(Credits_window,default_font,"          Brad Martin          ",0,default_font->get_heightspace());
 PlotString(Credits_window,default_font," Logo by: neptron              ",0,default_font->get_heightspace()*2);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
}

WINDOW *Screen_window=0;
#define NUM_SCREEN_OPTIONS 8
char *Screen_Options[NUM_SCREEN_OPTIONS]={
 "320x200x256 VGA",
 "320x200x256 VGA  FIT",
 "320x240x256 MODE-X",
 "256x239x256 VGA",
 "320x200x16b SVGA",
 "320x240x16b SVGA",
 "640x480x16b SVGA",
 "640x480x16b SVGA FIT"};

void UpdateScreenWindow(int Selected){
 Screen_window->refresh();
 for(int a=0;a<NUM_SCREEN_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Screen_window,default_font,Screen_Options[a],0,default_font->get_heightspace()*a,20);
  else PlotSelectedMenuItem(Screen_window,default_font,Screen_Options[a],0,default_font->get_heightspace()*a,20);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();
}

int ScreenWindow(){
 int CursorAt = PC_SCREEN_MODE;

 UpdateScreenWindow(CursorAt);

 for(;;){
  if(key[KEY_UP]){
   if(CursorAt) CursorAt--; else CursorAt=NUM_SCREEN_OPTIONS-1;
   UpdateScreenWindow(CursorAt);
   while (key[KEY_UP]);
  }
  if (key[KEY_DOWN]){
   if(CursorAt < NUM_SCREEN_OPTIONS-1) CursorAt++; else CursorAt=0;
   UpdateScreenWindow(CursorAt);
   while(key[KEY_DOWN]);
  }
  if(key[KEY_ESC]){
   while(key[KEY_ESC]);
   return -1;
  }
  if(key[KEY_ENTER]){
   while(key[KEY_ENTER]);
   return CursorAt;
  }
 }
 return -1; // Signify normal exit
}

char min_frameskip_str[] = "Min Frameskip: ???????";
char max_frameskip_str[] = "Max Frameskip: ???????";

WINDOW *Config_window=0;
#ifdef VAR_CYCLE_RATIO
#define NUM_CONFIG_OPTIONS 13
unsigned long SPC_CPU_cycle_divisor = 61, SPC_CPU_cycle_multiplicand = 6;
char spc_div_str[] = "SPC div: ???????";
char spc_mul_str[] = "SPC mul: ???????";
#else
#define NUM_CONFIG_OPTIONS 11
#endif
const char *Config_Options[NUM_CONFIG_OPTIONS]={
 0,                             // screen mode
 min_frameskip_str,
 max_frameskip_str,
 "-------- on player 1",        // controls: first player
 "-------- on player 2",        // controls: second player
 "Define keys for player 1",
 "Define keys for player 2",
 0,                             // SPC enable setting
 0,                             // Sound enable setting
 "ENVX reading: off",
 "FPS counter: off"
#ifdef VAR_CYCLE_RATIO
 "FPS counter: off",
 spc_div_str,
 spc_mul_str
#else
 "FPS counter: off"
#endif
};

char *on_off[2]={ "off", "on" };
char *en_dis[2]={ "enabled", "disabled" };

void UpdateConfigWindow(int Selected){
 Config_window->refresh();
 for(int a=0;a<NUM_CONFIG_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Config_window,default_font,Config_Options[a],0,default_font->get_heightspace()*a,24);
  else PlotSelectedMenuItem(Config_window,default_font,Config_Options[a],0,default_font->get_heightspace()*a,24);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();
}

WINDOW *ControlSetup_window=0;

void UpdateController1Screen(){
 char tempch[5];

 ControlSetup_window->refresh();

 scantotext(KEY_UP_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,31,15,8);
 scantotext(KEY_DOWN_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,65,15,8);
 scantotext(KEY_LEFT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,20,48,15,8);
 scantotext(KEY_RIGHT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,55,48,15,8);
 scantotext(KEY_A_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,151,46,15,8);
 scantotext(KEY_B_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,130,69,15,8);
 scantotext(KEY_X_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,25,15,8);
 scantotext(KEY_Y_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,104,49,15,8);
 scantotext(KEY_L_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,40,1,15,8);
 scantotext(KEY_R_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,1,15,8);
 scantotext(KEY_SELECT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,68,40,15,8);
 scantotext(KEY_START_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,91,40,15,8);
}

void UpdateController2Screen(){
 char tempch[5];

 ControlSetup_window->refresh();

 scantotext(KEY_UP_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,31,15,8);
 scantotext(KEY_DOWN_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,65,15,8);
 scantotext(KEY_LEFT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,20,48,15,8);
 scantotext(KEY_RIGHT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,55,48,15,8);
 scantotext(KEY_A_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,151,46,15,8);
 scantotext(KEY_B_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,130,69,15,8);
 scantotext(KEY_X_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,25,15,8);
 scantotext(KEY_Y_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,104,49,15,8);
 scantotext(KEY_L_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,40,1,15,8);
 scantotext(KEY_R_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,1,15,8);
 scantotext(KEY_SELECT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,68,40,15,8);
 scantotext(KEY_START_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,91,40,15,8);
}

int AskKey(char *msg, unsigned *whatkey, int jp){
 unsigned tmp;

 PlotStringShadow(ControlSetup_window,default_font,msg,84,96,15,8);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 CopyGUIScreen();

 tmp=WaitForKeyPress();

 if(tmp==KEY_ESC) return FALSE;

 *whatkey=tmp;

 if(jp==1) UpdateController1Screen();
 else UpdateController2Screen();

 return TRUE;
}

extern "C" unsigned char FPS_ENABLED;

int ConfigWindow(){
 int temp;

 Config_Options[0] = Screen_Options[PC_SCREEN_MODE];

 switch(JOYSTICK_ENABLED){
  case 1:
// Config_Options[3] = "Joystick on player 1";
// JOYSTICK_ENABLED=1;
// break;
  case 2:
   if(MOUSE_PLUGGED_IN){
    Config_Options[3] = "Mouse on player 1";
    JOYSTICK_ENABLED=2;
    break;
   }
  default:
   Config_Options[3] = "Keyboard on player 1";
   JOYSTICK_ENABLED=0;
 }

 switch(JOYSTICK_ENABLED2){
  case 1:
// Config_Options[4] = "Joystick on player 2";
// JOYSTICK_ENABLED2=1;
// break;
  case 2:
   if(MOUSE_PLUGGED_IN){
    Config_Options[4] = "Mouse on player 2";
    JOYSTICK_ENABLED2=2;
    break;
   }
  default:
   Config_Options[4] = "Keyboard on player 2";
   JOYSTICK_ENABLED2=0;
 }

 if(SPC_ENABLED) Config_Options[7] = "Emulate SPC";
 else Config_Options[7] = "Skip SPC";

 switch(SOUND_ENABLED){
  case 0:
   Config_Options[8] = "Sound Disabled";
   break;
  case 1:
   Config_Options[8] = "Sound Enabled (mono)";
   break;
  default:
   Config_Options[8] = "Sound Enabled (stereo)";
 }

 if (ENVX_ENABLED) Config_Options[9] = "ENVX reading: on";
 else Config_Options[9] = "ENVX reading: off";

 if (FPS_ENABLED) Config_Options[10] = "FPS counter: on";
 else Config_Options[10] = "FPS counter: off";

 sprintf(min_frameskip_str, "Min Frameskip: %-2ld", FRAME_SKIP_MIN);
 sprintf(max_frameskip_str, "Max Frameskip: %-2ld", FRAME_SKIP_MAX);
#ifdef VAR_CYCLE_RATIO
 sprintf(spc_div_str, "SPC div: %-5ld", SPC_CPU_cycle_divisor);
 sprintf(spc_mul_str, "SPC mul: %-5ld", SPC_CPU_cycle_multiplicand);
#endif

 UpdateConfigWindow(0);
 int CursorAt=0;

 for(;;){
  if(key[KEY_UP]){
   CursorAt--;
   if(CursorAt == -1) // so it wraps
    CursorAt = NUM_CONFIG_OPTIONS-1;
   UpdateConfigWindow(CursorAt);
    while(key[KEY_UP]);
  }
  if(key[KEY_DOWN]){
   CursorAt++;
   if(CursorAt == NUM_CONFIG_OPTIONS) // so it wraps
   CursorAt=0;
   UpdateConfigWindow(CursorAt);
   while(key[KEY_DOWN]);
  }

  // For min frameskip!
  if(CursorAt==1){
   if(key[KEY_LEFT] && FRAME_SKIP_MIN>0){
    // 60,30,20,15,10-0
    if(FRAME_SKIP_MIN>30) FRAME_SKIP_MIN-=30;
    else if(FRAME_SKIP_MIN>20) FRAME_SKIP_MIN-=10;
    else if(FRAME_SKIP_MIN>10) FRAME_SKIP_MIN-=5;
    else FRAME_SKIP_MIN--;

    if(FRAME_SKIP_MIN>FRAME_SKIP_MAX){
     FRAME_SKIP_MAX=FRAME_SKIP_MIN;
     sprintf(max_frameskip_str, "Max Frameskip: %-2ld", FRAME_SKIP_MAX);
    }
    sprintf(min_frameskip_str, "Min Frameskip: %-2ld", FRAME_SKIP_MIN);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && FRAME_SKIP_MIN<60){
    // 0-10,15,20,30,60
    if(FRAME_SKIP_MIN<10) FRAME_SKIP_MIN++;
    else if(FRAME_SKIP_MIN<20) FRAME_SKIP_MIN+=5;
    else if(FRAME_SKIP_MIN<30) FRAME_SKIP_MIN+=10;
    else FRAME_SKIP_MIN+=30;

    if(FRAME_SKIP_MIN>FRAME_SKIP_MAX){
     FRAME_SKIP_MAX=FRAME_SKIP_MIN;
     sprintf(max_frameskip_str, "Max Frameskip: %-2ld", FRAME_SKIP_MAX);
    }
    sprintf(min_frameskip_str, "Min Frameskip: %-2ld", FRAME_SKIP_MIN);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
  // For max frameskip!
  if(CursorAt==2){
   if(key[KEY_LEFT] && FRAME_SKIP_MAX>1){
    // 60,30,20,15,10-0
    if(FRAME_SKIP_MAX>30) FRAME_SKIP_MAX-=30;
    else if(FRAME_SKIP_MAX>20) FRAME_SKIP_MAX-=10;
    else if(FRAME_SKIP_MAX>10) FRAME_SKIP_MAX-=5;
    else FRAME_SKIP_MAX--;

    if(FRAME_SKIP_MAX<FRAME_SKIP_MIN){
     FRAME_SKIP_MIN=FRAME_SKIP_MAX;
     sprintf(min_frameskip_str, "Min Frameskip: %-2ld", FRAME_SKIP_MIN);
    }
    sprintf(max_frameskip_str, "Max Frameskip: %-2ld", FRAME_SKIP_MAX);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && FRAME_SKIP_MAX<60){
    // 0-10,15,20,30,60
    if(FRAME_SKIP_MAX<10) FRAME_SKIP_MAX++;
    else if(FRAME_SKIP_MAX<20) FRAME_SKIP_MAX+=5;
    else if(FRAME_SKIP_MAX<30) FRAME_SKIP_MAX+=10;
    else FRAME_SKIP_MAX+=30;

    if(FRAME_SKIP_MAX<FRAME_SKIP_MIN){
     FRAME_SKIP_MIN=FRAME_SKIP_MAX;
     sprintf(min_frameskip_str, "Min Frameskip: %-2ld", FRAME_SKIP_MIN);
    }
    sprintf(max_frameskip_str, "Max Frameskip: %-2ld", FRAME_SKIP_MAX);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
#ifdef VAR_CYCLE_RATIO
  if (CursorAt == 11){
   if(key[KEY_LEFT] && SPC_CPU_cycle_divisor>1){
    SPC_CPU_cycle_divisor--;
    sprintf(spc_div_str, "SPC div: %-5ld", SPC_CPU_cycle_divisor);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && SPC_CPU_cycle_divisor<99999){
    SPC_CPU_cycle_divisor++;
    sprintf(spc_div_str, "SPC div: %-5ld", SPC_CPU_cycle_divisor);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
  if (CursorAt == 12){
   if(key[KEY_LEFT] && SPC_CPU_cycle_multiplicand>1){
    SPC_CPU_cycle_multiplicand--;
    sprintf(spc_mul_str, "SPC mul: %-5ld", SPC_CPU_cycle_multiplicand);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && SPC_CPU_cycle_multiplicand<99999){
    SPC_CPU_cycle_multiplicand++;
    sprintf(spc_mul_str, "SPC mul: %-5ld", SPC_CPU_cycle_multiplicand);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
#endif
  if (key[KEY_ESC]){
   while(key[KEY_ESC]);
   return -1;
  }
  if(key[KEY_ENTER]){
   while(key[KEY_ENTER]);
   if(CursorAt==0){
    while((temp=ScreenWindow())!=-1){
     SetGUIScreen(temp);    // Setup screen mode (PC_SCREEN_MODE is set here so following works)

     Config_Options[0] = Screen_Options[PC_SCREEN_MODE];
     if(sneese){  // Prevent a crash if file not found!
      set_palette(sneesepal);
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     }
     set_palette_range(&GUIPal[-240],240,255,1);    // Set the GUI palette up.
     UpdateConfigWindow(0);
    }
    UpdateGUI(3);
   }
   if(CursorAt==3){
    switch(++JOYSTICK_ENABLED){
     case 1:
//    Config_Options[3] = "Joystick on player 1";
//    JOYSTICK_ENABLED=1;
//    break;
     case 2:
      if(MOUSE_PLUGGED_IN){
       Config_Options[3] = "Mouse on player 1";
       JOYSTICK_ENABLED=2;
       break;
      }
     default:
      Config_Options[3] = "Keyboard on player 1";
      JOYSTICK_ENABLED=0;
    }
   }
   if(CursorAt==4){
    switch(++JOYSTICK_ENABLED2){
     case 1:
//    Config_Options[4] = "Joystick on player 2";
//    JOYSTICK_ENABLED2=1;
//    break;
     case 2:
      if(MOUSE_PLUGGED_IN){
       Config_Options[4] = "Mouse on player 2";
       JOYSTICK_ENABLED2=2;
       break;
      }
     default:
      Config_Options[4] = "Keyboard on player 2";
      JOYSTICK_ENABLED2=0;
    }
   }
   if(CursorAt==5){
    while(key[KEY_ENTER]);
    clear_keybuf();
    UpdateController1Screen();

    if(AskKey("Press key for UP",&KEY_UP_1,1)==TRUE)
    if(AskKey("Press key for DOWN",&KEY_DOWN_1,1)==TRUE)
    if(AskKey("Press key for LEFT",&KEY_LEFT_1,1)==TRUE) // this is what I call optimization :)
    if(AskKey("Press key for RIGHT",&KEY_RIGHT_1,1)==TRUE)
    if(AskKey("Press key for A",&KEY_A_1,1)==TRUE)
    if(AskKey("Press key for B",&KEY_B_1,1)==TRUE)
    if(AskKey("Press key for X",&KEY_X_1,1)==TRUE)
    if(AskKey("Press key for Y",&KEY_Y_1,1)==TRUE)
    if(AskKey("Press key for L",&KEY_L_1,1)==TRUE)
    if(AskKey("Press key for R",&KEY_R_1,1)==TRUE)
    if(AskKey("Press key for SELECT",&KEY_SELECT_1,1)==TRUE)
    if(AskKey("Press key for START",&KEY_START_1,1)==TRUE)
    {
     PlotStringShadow(ControlSetup_window,default_font,"Press ESC to exit",84,96,15,8);
     if(sneese)
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
     CopyGUIScreen();

     while (!key[KEY_ESC]);   // :)
    }
    while (key[KEY_ESC]);

    UpdateGUI(3);
    UpdateConfigWindow(CursorAt);
   }
   if(CursorAt==6){
    while(key[KEY_ENTER]);
    clear_keybuf();
    UpdateController2Screen();

    if(AskKey("Press key for UP",&KEY_UP_2,2)==TRUE)
    if(AskKey("Press key for DOWN",&KEY_DOWN_2,2)==TRUE)
    if(AskKey("Press key for LEFT",&KEY_LEFT_2,2)==TRUE) // this is what I call optimization :)
    if(AskKey("Press key for RIGHT",&KEY_RIGHT_2,2)==TRUE)
    if(AskKey("Press key for A",&KEY_A_2,2)==TRUE)
    if(AskKey("Press key for B",&KEY_B_2,2)==TRUE)
    if(AskKey("Press key for X",&KEY_X_2,2)==TRUE)
    if(AskKey("Press key for Y",&KEY_Y_2,2)==TRUE)
    if(AskKey("Press key for L",&KEY_L_2,2)==TRUE)
    if(AskKey("Press key for R",&KEY_R_2,2)==TRUE)
    if(AskKey("Press key for SELECT",&KEY_SELECT_2,2)==TRUE)
    if(AskKey("Press key for START",&KEY_START_2,2)==TRUE)
    {
     PlotStringShadow(ControlSetup_window,default_font,"Press ESC to exit",84,96,15,8);
     if(sneese)
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
     CopyGUIScreen();

     while(!key[KEY_ESC]);
    }
    while (key[KEY_ESC]);

    UpdateGUI(3);
    UpdateConfigWindow(CursorAt);
   }
   if(CursorAt==7){
    if(SPC_ENABLED){
     Config_Options[7] = "Skip SPC";
     SPC_ENABLED=0;
    } else {
     Config_Options[7] = "Emulate SPC";
     SPC_ENABLED=1;
    }
   }
   if(CursorAt==8){
    switch(SOUND_ENABLED){
     case 0:
      if(Install_Sound(0)){
       Config_Options[8] = "Sound Enabled (mono)";
      }
      break;
     case 1:
      Remove_Sound();
      if(Install_Sound(1)){
       Config_Options[8] = "Sound Enabled (stereo)";
      } else {
       Config_Options[8] = "Sound Disabled";
      }
      break;
     case 2:
      Remove_Sound();
      Config_Options[8] = "Sound Disabled";
    }
   }
   if (CursorAt == 9){
    Config_Options[9] = (ENVX_ENABLED ^= -1) ?
     "ENVX reading: on" : "ENVX reading: off";
   }
   if (CursorAt == 10){
    Config_Options[10] = (FPS_ENABLED ^= -1) ?
     "FPS counter: on" : "FPS counter: off";
   }
   UpdateConfigWindow(CursorAt);
  }
 }
 return 1;          // Signify normal exit
}

WINDOW *ROMInfo_window=0;

void RomInfo(void){
 ROMInfo_window->refresh();
 PlotString(ROMInfo_window,default_font,rom_romfile,0,0);
 PlotString(ROMInfo_window,default_font,rom_romname,0,default_font->get_heightspace());
 PlotString(ROMInfo_window,default_font,rom_romtype,0,default_font->get_heightspace()*2);
 PlotString(ROMInfo_window,default_font,rom_romhilo,0,default_font->get_heightspace()*3);
 PlotString(ROMInfo_window,default_font,rom_romsize,0,default_font->get_heightspace()*4);
 PlotString(ROMInfo_window,default_font,rom_sram,0,default_font->get_heightspace()*5);
 PlotString(ROMInfo_window,default_font,rom_country,0,default_font->get_heightspace()*6);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);   // :)
 while(key[KEY_ESC]);
}

extern unsigned char BGMODE,VMAIN;
extern "C" unsigned char W12SEL,W34SEL,WOBJSEL,WH0,WH1,WH2,WH3;
extern "C" unsigned char TM,TS,TMW,TSW,WBGLOG,WOBJLOG,CGWSEL,CGADSUB;

WINDOW *HWStatus_window=0;

void HWStatus(void){
 char Number[5];

 HWStatus_window->refresh();
 PlotString(HWStatus_window,default_font,"65c816 A:#### X:#### Y:#### DB:##",0,0);
 sprintf(Number,"%04X",(unsigned short) _65816_A);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*9,0);
 sprintf(Number,"%04X",(unsigned short) _65816_X);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,0);
 sprintf(Number,"%04X",(unsigned short) _65816_Y);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*23,0);
 sprintf(Number,"%02X",(unsigned short) _65816_DB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*31,0);
 PlotString(HWStatus_window,default_font,"PB:## PC:#### S:#### D:#### P:###",0,default_font->get_heightspace());
 sprintf(Number,"%02X",(unsigned short) _65816_PB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*3,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_PC);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*9,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_S);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_D);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*23,default_font->get_heightspace());
// sprintf(Number,"%03X",_65816_P&0x1FF
// PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*30,default_font->get_heightspace());
 PlotString(HWStatus_window,default_font,"OLD_PB:## OLD_PC:####",0,default_font->get_heightspace()*2);
 sprintf(Number,"%02X",(unsigned char) OLD_PB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*7,default_font->get_heightspace()*2);
 sprintf(Number,"%04X",(unsigned short) OLD_PC);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*17,default_font->get_heightspace()*2);

 PlotString(HWStatus_window,default_font,"BGMODE:## VMAIN:## TM:## TS:##",0,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",BGMODE);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*7,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",VMAIN);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",TM);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*22,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",TS);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*28,default_font->get_heightspace()*3);
 PlotString(HWStatus_window,default_font,"WH0: ## WH1: ## WH2: ## WH3: ##",0,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH0);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH1);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*13,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH2);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*21,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH3);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*29,default_font->get_heightspace()*4);
 PlotString(HWStatus_window,default_font,"TMW: ## TSW: ## WSEL: ## ## ##",0,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",TMW);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",TSW);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*13,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",W12SEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*22,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",W34SEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*25,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",WOBJSEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*28,default_font->get_heightspace()*5);

 PlotString(HWStatus_window,default_font,"WLOG:## ## CGWSEL:## CGADSUB:##",0,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",WBGLOG);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",WOBJLOG);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*8,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",CGWSEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*18,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",CGADSUB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*29,default_font->get_heightspace()*6);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);
 while(key[KEY_ESC]);
}

extern "C" unsigned char MDMAEN;
extern "C" unsigned char HDMAEN;
extern "C" unsigned char DMAP_0, DMAP_1, DMAP_2, DMAP_3;
extern "C" unsigned char DMAP_4, DMAP_5, DMAP_6, DMAP_7;
extern "C" unsigned char BBAD_0, BBAD_1, BBAD_2, BBAD_3;
extern "C" unsigned char BBAD_4, BBAD_5, BBAD_6, BBAD_7;
extern "C" unsigned char A1TL_0, A1TL_1, A1TL_2, A1TL_3;
extern "C" unsigned char A1TL_4, A1TL_5, A1TL_6, A1TL_7;
extern "C" unsigned char A1TH_0, A1TH_1, A1TH_2, A1TH_3;
extern "C" unsigned char A1TH_4, A1TH_5, A1TH_6, A1TH_7;
extern "C" unsigned char A1B_0, A1B_1, A1B_2, A1B_3;
extern "C" unsigned char A1B_4, A1B_5, A1B_6, A1B_7;
extern "C" unsigned char DASL_0, DASL_1, DASL_2, DASL_3;
extern "C" unsigned char DASL_4, DASL_5, DASL_6, DASL_7;
extern "C" unsigned char DASH_0, DASH_1, DASH_2, DASH_3;
extern "C" unsigned char DASH_4, DASH_5, DASH_6, DASH_7;
extern "C" unsigned char DASB_0, DASB_1, DASB_2, DASB_3;
extern "C" unsigned char DASB_4, DASB_5, DASB_6, DASB_7;
extern "C" unsigned char A2L_0, A2L_1, A2L_2, A2L_3;
extern "C" unsigned char A2L_4, A2L_5, A2L_6, A2L_7;
extern "C" unsigned char A2H_0, A2H_1, A2H_2, A2H_3;
extern "C" unsigned char A2H_4, A2H_5, A2H_6, A2H_7;
extern "C" unsigned char A2B_0, A2B_1, A2B_2, A2B_3;
extern "C" unsigned char A2B_4, A2B_5, A2B_6, A2B_7;

WINDOW *DMAStatus_window=0;

void DMAStatus(void){
 char Line[34];

 DMAStatus_window->refresh();
 sprintf(Line,"MDMAEN:%02X HDMAEN:%02X",(unsigned char) MDMAEN,
  (unsigned char) HDMAEN);
 PlotString(DMAStatus_window,default_font,Line,0,0);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_0,(unsigned char) BBAD_0,
 (unsigned char) A1B_0,(unsigned char) A1TH_0,(unsigned char) A1TL_0,
 (unsigned char) A2B_0,(unsigned char) A2H_0,(unsigned char) A2L_0,
 (unsigned char) DASB_0,(unsigned char) DASH_0,(unsigned char) DASL_0);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace());

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_1,(unsigned char) BBAD_1,
 (unsigned char) A1B_1,(unsigned char) A1TH_1,(unsigned char) A1TL_1,
 (unsigned char) A2B_1,(unsigned char) A2H_1,(unsigned char) A2L_1,
 (unsigned char) DASB_1,(unsigned char) DASH_1,(unsigned char) DASL_1);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*2);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_2,(unsigned char) BBAD_2,
 (unsigned char) A1B_2,(unsigned char) A1TH_2,(unsigned char) A1TL_2,
 (unsigned char) A2B_2,(unsigned char) A2H_2,(unsigned char) A2L_2,
 (unsigned char) DASB_2,(unsigned char) DASH_2,(unsigned char) DASL_2);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*3);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_3,(unsigned char) BBAD_3,
 (unsigned char) A1B_3,(unsigned char) A1TH_3,(unsigned char) A1TL_3,
 (unsigned char) A2B_3,(unsigned char) A2H_3,(unsigned char) A2L_3,
 (unsigned char) DASB_3,(unsigned char) DASH_3,(unsigned char) DASL_3);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*4);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_4,(unsigned char) BBAD_4,
 (unsigned char) A1B_4,(unsigned char) A1TH_4,(unsigned char) A1TL_4,
 (unsigned char) A2B_4,(unsigned char) A2H_4,(unsigned char) A2L_4,
 (unsigned char) DASB_4,(unsigned char) DASH_4,(unsigned char) DASL_4);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*5);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_5,(unsigned char) BBAD_5,
 (unsigned char) A1B_5,(unsigned char) A1TH_5,(unsigned char) A1TL_5,
 (unsigned char) A2B_5,(unsigned char) A2H_5,(unsigned char) A2L_5,
 (unsigned char) DASB_5,(unsigned char) DASH_5,(unsigned char) DASL_5);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*6);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_6,(unsigned char) BBAD_6,
 (unsigned char) A1B_6,(unsigned char) A1TH_6,(unsigned char) A1TL_6,
 (unsigned char) A2B_6,(unsigned char) A2H_6,(unsigned char) A2L_6,
 (unsigned char) DASB_6,(unsigned char) DASH_6,(unsigned char) DASL_6);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*7);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP_7,(unsigned char) BBAD_7,
 (unsigned char) A1B_7,(unsigned char) A1TH_7,(unsigned char) A1TL_7,
 (unsigned char) A2B_7,(unsigned char) A2H_7,(unsigned char) A2L_7,
 (unsigned char) DASB_7,(unsigned char) DASH_7,(unsigned char) DASL_7);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*8);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);
 while(key[KEY_ESC]);
}

CTL *joypad_bitmap=0;

CTL_CLEAR GUI_clear;
CTL_CLEAR Main_clear;
CTL_CLEAR Credits_clear;
CTL_CLEAR Screen_clear;
CTL_CLEAR Config_clear;
CTL_CLEAR ControlSetup_clear;
CTL_CLEAR ROMInfo_clear;
CTL_CLEAR HWStatus_clear;
CTL_CLEAR DMAStatus_clear;
CTL_CLEAR File_clear;

int allocate_windows(){
 static int file_cleared=0;
 if(!file_cleared) File_window+=File_clear;

 GUI_window=new BORDER_WINDOW(-1,-1,320,240);
 Main_window=new BORDER_WINDOW(0,0,16*default_font->get_widthspace(),NUM_OPTIONS*default_font->get_heightspace());
 Credits_window=new BORDER_WINDOW(10,160,32*default_font->get_widthspace(),3*default_font->get_heightspace());
 Screen_window=new BORDER_WINDOW(100,30,20*default_font->get_widthspace(),NUM_SCREEN_OPTIONS*default_font->get_heightspace());
 Config_window=new BORDER_WINDOW(70,15,24*default_font->get_widthspace(),NUM_CONFIG_OPTIONS*default_font->get_heightspace());
 ControlSetup_window=new BORDER_WINDOW(29,34,168,104);
 ROMInfo_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),7*default_font->get_heightspace());
 HWStatus_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),7*default_font->get_heightspace());
 DMAStatus_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),9*default_font->get_heightspace());
 if(!(GUI_window && Main_window && Credits_window &&
    Screen_window && Config_window && ControlSetup_window &&
    ROMInfo_window && HWStatus_window && DMAStatus_window)) return 1;
 *GUI_window+=GUI_clear;
 *Main_window+=Main_clear;
 *Credits_window+=Credits_clear;
 *Screen_window+=Screen_clear;
 *Config_window+=Config_clear;
 *ControlSetup_window+=ControlSetup_clear;
 if(joypad)
  joypad_bitmap=new CTL_BITMAP(ControlSetup_window,joypad,7,12);
 *ROMInfo_window+=ROMInfo_clear;
 *HWStatus_window+=HWStatus_clear;
 *DMAStatus_window+=DMAStatus_clear;
 return 0;
}

void free_windows(){
 delete GUI_window;
 delete Main_window; delete Credits_window;
 delete joypad_bitmap;
 delete Screen_window; delete Config_window; delete ControlSetup_window;
 delete ROMInfo_window; delete HWStatus_window; delete DMAStatus_window;

 GUI_window=0;
 Main_window=0; Credits_window=0;
 joypad_bitmap=0;
 Screen_window=0; Config_window=0; ControlSetup_window=0;
 ROMInfo_window=0; HWStatus_window=0; DMAStatus_window=0;
}

extern volatile unsigned Timer_Counter_Throttle;
extern volatile unsigned Timer_Counter_Profile;
/*
extern volatile unsigned Timer_Counter_FPS;
 */

int GUI(){

#ifdef DEBUG
 unsigned long Timer_Value,Timer_Value2;
/*
 unsigned long Timer_Value3;
 */
#endif
 unsigned char Old_SPC=SPC_ENABLED;

#ifdef DEBUG
 Timer_Value=Timer_Counter_Profile;
 Timer_Value2=Timer_Counter_Throttle;
/*
 Timer_Value3=Timer_Counter_FPS;
 */
#endif

 allocate_windows();

 if(sneese){ // Prevent a crash if file not found!
  set_palette(sneesepal);
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 }
 while (key[KEY_ESC]);

 char TempBufferP[MAXPATH]; // For saving the current directory to use!

 set_palette_range(&GUIPal[-240],240,255,1);    // Set the GUI palette up.

 UpdateGUI(0);
 CopyGUIScreen();

 int CursorAt=0;

 for(;;){
  if(key[KEY_UP]){
   CursorAt--;
   if(CursorAt == -1) // so it wraps
    CursorAt = NUM_OPTIONS-1;
   UpdateGUI(CursorAt);
   CopyGUIScreen();
   while(key[KEY_UP]);
  }
  if(key[KEY_DOWN]){
   CursorAt++;
   if(CursorAt == NUM_OPTIONS) // so it wraps
    CursorAt=0;
   UpdateGUI(CursorAt);
   CopyGUIScreen();
   while (key[KEY_DOWN]);
  }
  if(key[KEY_ESC]){
   if(ROM_Loaded){
    while(key[KEY_ESC]);
    break;
   }
  }
  if(key[KEY_ENTER]){
   while (key[KEY_ENTER]);
   if(ROM_Loaded){
    if(CursorAt==0) break;
    if(CursorAt==1){
     Reset_Memory();
     Reset_CPU();
     Reset_SPC();
     break;
    }
   }
   if(CursorAt==2){
    getcwd(TempBufferP,MAXPATH);
    if(FileWindow()==0){
     chdir(TempBufferP);
     break;
    }
    chdir(TempBufferP);
   }
   if(CursorAt==3)
    ConfigWindow();
   if(ROM_Loaded){
    if(CursorAt==4)
     RomInfo();
    if(CursorAt==5)
     HWStatus();
    if(CursorAt==6)
     DMAStatus();
   }
   if(CursorAt==7){
#ifdef DEBUG
    Timer_Counter_Profile=Timer_Value;
    Timer_Counter_Throttle=Timer_Value2;
/*
    Timer_Counter_FPS=Timer_Value3;
 */
#endif
    free_windows();
    return 0;
   }
   UpdateGUI(CursorAt);
   CopyGUIScreen();
  }
 }

 if(ROM_Loaded && SPC_ENABLED && (SPC_ENABLED!=Old_SPC)){
  Reset_Memory();
  Reset_CPU();          // Must be done to ensure ROM works ok!
  Reset_SPC();
 }

#ifdef DEBUG
 Timer_Counter_Profile=Timer_Value;
 Timer_Counter_Throttle=Timer_Value2;
/*
 Timer_Counter_FPS=Timer_Value2;
 */
#endif

 clear(screen);
 clear(Allegro_Bitmap);

 free_windows();
 return 1;          // Signify normal exit
}
