;%define TRAP_BAD_READS
;%define TRAP_BAD_WRITES
;%define TRAP_IGNORED_WRITES
;
;
; SNEeSe SNES memory mapper v1.6 - NASM rewrite
;
; There are numerous jump tables in this code, but a lot less than before...
;  back down from 4k to 256... easier on the cache this way, and with
;  dynamic memory map init both LoROM and HiROM are very simply mapped...
;
;  Files are split up as follows:
;
;    mem.S      - LoROM memory map code/tables.
;    memhi.S    - HiROM memory map code/tables.
;    memmap.i   - Memory mapping macros.
;    memblank.S - 'Blank' ROM bank handlers for dynamic memory map
;                 and code common to both LoROM and HiROM.
;    memhard.S  - Hardware port (21xx/40xx/42xx/43xx) handlers.
;    memDMA.S   - DMA emulation code.
;
;

%include "misc.ni"
%include "memmap.ni"

EXPORT_C memblank_start

extern _SRAM_READ
extern _WRAM,_SRAM_Mask,_SRAM
extern Read_Map_20_5F,Write_Map_20_5F

section .data
ALIGND
EXPORT_C LoROM_Blank
; Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_00_3F

; Banks 40-6F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,48,_BlankL_40_6F

; Banks 70-7D is SRAM (mirrored)
    DUPLICATE dd,14,_SRAM_READ

; Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_READ

; Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_80_BF

; Banks C0-FF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE dd,64,_BlankL_C0_FF

ALIGND
EXPORT_C HiROM_Blank
; Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_BlankH_00_3F_No_SRAM

; Banks 40-7D are 0000-FFFF ROM
    DUPLICATE dd,62,_BlankH_40_7D

; Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE dd,2,_RAM_READ

; Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
; Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE dd,64,_BlankH_80_BF_No_SRAM

; Banks C0-FF are 0000-FFFF ROM
    DUPLICATE dd,64,_BlankH_C0_FF

section .bss
ALIGND
EXPORT_C Read_BankMapping   ,skipl 256

ALIGND
EXPORT_C Write_BankMapping  ,skipl 256

ALIGND
EXPORT_C Read_BankOffset    ,skipl 256*2

ALIGND
EXPORT_C Read_Bank8Mapping  ,skipl 256*8

ALIGND
EXPORT_C Write_Bank8Mapping ,skipl 256*8

ALIGND
EXPORT_C Read_Bank8Offset   ,skipl 256*8

ALIGND
EXPORT_C Write_Bank8Offset  ,skipl 256*8

ALIGND
EXPORT_C Dummy              ,skipk 64

EXPORT_C RomReadAdjust      ,skipl
EXPORT_C RomReadAdjust2     ,skipl
EXPORT_C RomHiReadAdjust    ,skipl
EXPORT_C RomHiReadAdjust2   ,skipl
EXPORT_C HRomLinReadAdjust  ,skipl
EXPORT_C HRomHiLinReadAdjust,skipl

extern _Map_Address
extern _Map_Byte

section .text
ALIGNC
EXPORT UNSUPPORTED_READ
EXPORT_C UNSUPPORTED_READ
    mov al,0
%ifdef DEBUG
%ifdef TRAP_BAD_READS
extern _InvalidHWRead
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWRead     ; Unmapped hardware address!
    popa
%endif
%endif
    ret

ALIGNC
EXPORT UNSUPPORTED_WRITE
EXPORT_C UNSUPPORTED_WRITE
%ifdef DEBUG
%ifdef TRAP_BAD_WRITES
extern _InvalidHWWrite
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWWrite    ; Unmapped hardware address!
    popa
%endif
%endif
    ret

ALIGNC
EXPORT IGNORE_WRITE
EXPORT_C IGNORE_WRITE
%ifdef DEBUG
%ifdef TRAP_IGNORED_WRITES
extern _InvalidHWWrite
    mov [_Map_Address],ebx  ; Set up Map Address so message works!
    mov [_Map_Byte],al      ; Set up Map Byte so message works
    pusha
    call _InvalidHWWrite    ; Unmapped hardware address!
    popa
%endif
%endif
    ret

ALIGNC
EXPORT_C RAM_READ
    mov al,[_WRAM-0x7E0000+ebx]
    ret

ALIGNC
EXPORT_C RAM_WRITE
    mov [_WRAM-0x7E0000+ebx],al
    ret

ALIGNC
EXPORT_C READ_00_3F_No_SRAM
    test bh,bh
    jns short .not_rom
    ROM00_READ_32k
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

ALIGNC
EXPORT_C READ_80_BF_No_SRAM
    test bh,bh
    jns short .not_rom
    ROM80_READ_32k
ALIGNC
.not_rom:
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    NON_MAPPED_READ

ALIGNC
EXPORT_C BlankL_00_3F
EXPORT_C BlankL_80_BF
EXPORT_C BlankH_00_3F_No_SRAM
EXPORT_C BlankH_80_BF_No_SRAM
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    jns short .not_rom
    NO_ROM_READ
ALIGNC
.not_rom:
    NON_MAPPED_READ

ALIGNC
EXPORT_C BlankL_40_6F
EXPORT_C BlankL_C0_FF
    test bh,bh
    jns short .not_rom
    NO_ROM_READ
ALIGNC
.not_rom:
    NON_MAPPED_READ

%macro _BlankH_30_3F 1  ;bank
ALIGNC
EXPORT_C BlankH_%1
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    js  short .not_sram
    SRAM30_READ %1
ALIGNC
.not_sram:
    NO_ROM_READ
%endmacro

%macro _BlankH_B0_BF 1  ;bank
ALIGNC
EXPORT_C BlankH_%1
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_READ
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_READ
ALIGNC
.not_ppu:
    test bh,bh
    js  short .not_sram
    SRAMB0_READ %1
ALIGNC
.not_sram:
    NO_ROM_READ
%endmacro

ALIGNC
EXPORT_C BlankH_40_7D
EXPORT_C BlankH_C0_FF
    NO_ROM_READ

    _BlankH_30_3F 30
    _BlankH_30_3F 31
    _BlankH_30_3F 32
    _BlankH_30_3F 33
    _BlankH_30_3F 34
    _BlankH_30_3F 35
    _BlankH_30_3F 36
    _BlankH_30_3F 37
    _BlankH_B0_BF B0
    _BlankH_B0_BF B1
    _BlankH_B0_BF B2
    _BlankH_B0_BF B3
    _BlankH_B0_BF B4
    _BlankH_B0_BF B5
    _BlankH_B0_BF B6
    _BlankH_B0_BF B7

ALIGNC
EXPORT_C WRITE_00_3F            ; Used in LoROM!
EXPORT_C WRITE_80_BF
EXPORT_C HWRITE_00_3F_No_SRAM   ; Used in HiROM!
EXPORT_C HWRITE_80_BF_No_SRAM
    cmp bh,0x20
    jnb short .not_lowram
    LOW_RAM_WRITE
ALIGNC
.not_lowram:
    cmp bh,0x60
    jnb short .not_ppu
    PPU_WRITE
ALIGNC
.not_ppu:
    ROM_WRITE

ALIGNC
EXPORT_C PPU_READ
    PPU_READ

ALIGNC
EXPORT_C PPU_WRITE
    PPU_WRITE


section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
