
; Mode 7 matrix rendering / hardware port emulation.

%include "misc.ni"
%include "sprites.ni"
%include "screen.ni"

EXPORT_C mode7_start

extern Current_Line_Render,Current_Line_Mosaic
extern Ready_Line_Render,Sync_Render
extern MOSAIC,Mosaic_Size,MosaicBG1

extern ScreenDestPtr
extern _VRAM_DI
extern Tile_priority_opcode
extern Redo_OAM,Recache_OAM
extern NO_PLANES
extern _SETINI
extern SpriteList,NextSprite,Sprite_Mode
extern Priority_Used,Plot_Sprites

section .data
ALIGND
EXPORT M7_Table
dd M7_REPEAT,M7_REPEAT_X,M7_REPEAT_Y,M7_REPEAT_XY
dd M7_CLIP,M7_CLIP_X,M7_CLIP_Y,M7_CLIP_XY
dd M7_CLIP,M7_CLIP_X,M7_CLIP_Y,M7_CLIP_XY
dd M7_CHAR0,M7_CHAR0_X,M7_CHAR0_Y,M7_CHAR0_XY
dd M7P_REPEAT,M7P_REPEAT_X,M7P_REPEAT_Y,M7P_REPEAT_XY
dd M7P_CLIP,M7P_CLIP_X,M7P_CLIP_Y,M7P_CLIP_XY
dd M7P_CLIP,M7P_CLIP_X,M7P_CLIP_Y,M7P_CLIP_XY
dd M7P_CHAR0,M7P_CHAR0_X,M7P_CHAR0_Y,M7P_CHAR0_XY

section .bss
ALIGND
EXPORT Mode7_A_P,skipl  ; Partially precalculated
EXPORT Mode7_B  ,skipl
EXPORT Mode7_C_P,skipl  ; Partially precalculated
EXPORT Mode7_D  ,skipl

EXPORT_C M7A    ,skipl
EXPORT_C M7B    ,skipl
EXPORT_C M7C    ,skipl
EXPORT_C M7D    ,skipl
EXPORT_C M7X_13 ,skipl
EXPORT_C M7Y_13 ,skipl
EXPORT_C M7H    ,skipl
EXPORT_C M7V    ,skipl
EXPORT_C M7X    ,skipl
EXPORT_C M7Y    ,skipl

MPYL:   skipb   ; Mode 7 multiplication result: low byte
MPYM:   skipb   ; Mode 7 multiplication result: middle byte
MPYH:   skipb   ; Mode 7 multiplication result: high byte
        skipb

EXPORT M7_Handler,skipl
EXPORT_C M7SEL,skipb    ; ab0000yx  ab=mode 7 repetition info,y=flip vertical,x=flip horizontal
EXPORT Redo_M7,skipb
Redo_16x8:  skipb

section .text
ALIGNC
EXPORT RENDER_MODE_7
 CALCULATE_MOSAIC_YLINE 0x01

 mov edi,[ScreenDestPtr]

 mov dl,[Redo_M7]
 test dl,0x0F   ; Need to do any recalculating?
 jz .end_recalc

 test dl,0x0A   ; Recalculate B or D?
 jz .end_recalc_bd

 mov eax,[_M7V]
 sub eax,[_M7Y_13]
 test dl,0x02   ; Recalculate B?
 jz .recalc_d

 push eax
 imul eax,[_M7B]
 mov [Mode7_B],eax
 pop eax
 test dl,0x08   ; Recalculate D?
 jz .end_recalc_bd

.recalc_d:
 imul eax,[_M7D]
 mov [Mode7_D],eax

.end_recalc_bd:
 test dl,0x05   ; Recalculate A or C?
 jz .end_recalc

.no_recalc_bd:
 mov eax,[_M7H]
 sub eax,[_M7X_13]
 test dl,0x01   ; Recalculate A?
 jz .recalc_c

 push eax
 imul eax,[_M7A]
 mov ebx,[_M7X_13]
 shl ebx,8
 add eax,ebx
 mov [Mode7_A_P],eax
 pop eax
 test dl,0x04   ; Recalculate C?
 jz  .end_recalc

.recalc_c:
 imul eax,[_M7C]
 mov ebx,[_M7Y_13]
 shl ebx,8
 add eax,ebx
 mov [Mode7_C_P],eax

.end_recalc:
 mov byte [Redo_M7],0   ; Done with recalculating

 mov al,[MosaicBG1]
 test al,al
 mov ebx,[Current_Line_Render]
 jz .no_mosaic
 mov ebx,[Current_Line_Mosaic]

.no_mosaic:
 mov esi,_VRAM_DI
 mov ebp,256            ; Horizontal Count
 mov eax,[_M7B]
 mov ecx,[_M7A]
 mov edx,[_M7C]
 jmp dword [M7_Handler] ; New for 0.25

;%1 = xflip, %2 = priority, %3 = label, %4 = ylabel
%macro M7__REPEAT 4
ALIGNC
%4:
 xor bl,0xFF
%3:
%if %1
 add edi,255
%endif
 imul eax,ebx
 imul ebx,[_M7D]
 add eax,[Mode7_A_P]
 add ebx,[Mode7_C_P]
 add eax,[Mode7_B]
%if %2
 mov cl,[Tile_priority_opcode]
%endif
 add ebx,[Mode7_D]
%if %2
 mov [.priority_%3],cl
%endif

ALIGNC
.pixel_loop:
 cmp byte [edi],0
 jnz .pixel_covered

 push eax
 push ebx

 shr eax,8
 shr ebx,8
 and eax,0x3FF
 and ebx,0x3FF

; Convert Screen X,Y location to SNES Pic location

                ; Assumes %ax is X coord 0-1023, %bx is Y 0-1023
 mov ecx,eax
 mov edx,ebx
 shr eax,3      ; Get Tile Position (in 128 by 128 map)
 and ecx,byte 7 ; Get pixel shift within tile
 shl ebx,5      ; Tile Position*256 cos thats width of map
 and edx,byte 7
 add al,al
 mov bl,al      ; Map offset now in %ebx
 shr ebx,byte 1
 xor eax,eax
 shl edx,3
 mov al,[esi+ebx]   ; Got Tile Number
    
 shl eax,6          ; Get offset to tile data
 add eax,ecx        ; Add X offset
 add eax,edx        ; Add Y offset

 mov al,[32768+esi+eax]

%if %2
 test al,al
.priority_%3:
 js .wrong_priority
 and al,0x7F
%endif
 mov [edi],al

.wrong_priority:
 pop ebx
 pop eax
 mov ecx,[_M7A]
 mov edx,[_M7C]

.pixel_covered:
 add eax,ecx
 add ebx,edx

%if %1
 dec edi
%else
 inc edi
%endif

 dec ebp
 jnz .pixel_loop

 ret
%endmacro

; New for v0.16, tile 0 repeat support
;%1 = xflip, %2 = priority, %3 = label, %4 = ylabel
%macro M7__CHAR0 4
ALIGNC
%4:
 xor bl,0xFF
%3:
%if %1
 add edi,255
%endif
 imul eax,ebx
 imul ebx,[_M7D]
 add eax,[Mode7_A_P]
 add ebx,[Mode7_C_P]
 add eax,[Mode7_B]
%if %2
 mov cl,[Tile_priority_opcode]
%endif
 add ebx,[Mode7_D]
%if %2
 mov [.priority_%3],cl
%endif

ALIGNC
.pixel_loop:
 cmp byte [edi],0
 jnz .pixel_covered

 push eax
 push ebx

 shr eax,8
 shr ebx,8

; Convert Screen X,Y location to SNES Pic location

                ; Assumes %ax is X coord 0-1023, %bx is Y 0-1023
 mov ecx,eax
 mov edx,ebx
 and ecx,byte 7 ; Get pixel shift within tile
 and edx,byte 7

 cmp eax,1023   ; If outside screen range we use tile 0
 ja .use_tile_0
 cmp ebx,1023
 jbe .get_tile
.use_tile_0:
 xor eax,eax
 jmp .have_tile

ALIGNC
.get_tile:
 shr eax,3      ; Get Tile Position (in 128 by 128 map)
 shl ebx,5      ; Tile Position*256 cos thats width of map
 add al,al
 mov bl,al      ; Map offset now in %ebx
 shr ebx,byte 1
 xor eax,eax
 mov al,[esi+ebx]   ; Got Tile Number
    
 shl eax,6          ; Get offset to tile data
.have_tile:
 add eax,ecx        ; Add X offset
 shl edx,3
 add eax,edx        ; Add Y offset

 mov al,[32768+esi+eax]

%if %2
 test al,al
.priority_%3:
 js  .wrong_priority
 and al,0x7F
%endif
 mov [edi],al

.wrong_priority:
 pop ebx
 pop eax
 mov ecx,[_M7A]
 mov edx,[_M7C]

.pixel_covered:
 add eax,ecx
 add ebx,edx

%if %1
 dec edi
%else
 inc edi
%endif

 dec ebp
 jnz .pixel_loop

 ret
%endmacro

;%1 = xflip, %2 = priority, %3 = label, %4 = ylabel
%macro M7__CLIP 4
ALIGNC
%4:
 xor bl,0xFF
%3:
%if %1
 add edi,255
%endif
 imul eax,ebx
 imul ebx,[_M7D]
 add eax,[Mode7_A_P]
 add ebx,[Mode7_C_P]
 add eax,[Mode7_B]
%if %2
 mov cl,[Tile_priority_opcode]
%endif
 add ebx,[Mode7_D]
%if %2
 mov [.priority_%3],cl
%endif

ALIGNC
.pixel_loop:
 cmp byte [edi],0
 jnz .pixel_covered

 test eax,~0x3FFFF  ; If outside screen range we simply skip the pixel
 jnz .pixel_covered

 test ebx,~0x3FFFF
 jnz .pixel_covered

; Convert Screen X,Y location to SNES Pic location

                ; Assumes %ax is X coord 0-1023, %bx is Y 0-1023
 push eax
 push ebx
 shr eax,8
 shr ebx,8
 mov ecx,eax
 mov edx,ebx
 shr eax,3      ; Get Tile Position (in 128 by 128 map)
 and ecx,byte 7 ; Get pixel shift within tile
 shl ebx,5      ; Tile Position*256 cos thats width of map
 and edx,byte 7
 add al,al
 mov bl,al      ; Map offset now in %ebx
 shr ebx,byte 1
 xor eax,eax
 shl edx,3
 mov al,[esi+ebx]   ; Got Tile Number
    
 shl eax,6          ; Get offset to tile data
 add eax,ecx        ; Add X offset
 add eax,edx        ; Add Y offset

 mov al,[32768+esi+eax]
%if %2
 test al,al
.priority_%3:
 js .wrong_priority
 and al,0x7F
%endif
 mov [edi],al

.wrong_priority:
 pop ebx
 pop eax
 mov ecx,[_M7A]
 mov edx,[_M7C]

.pixel_covered:
 add eax,ecx
 add ebx,edx

%if %1
 dec edi
%else
 inc edi
%endif

 dec ebp
 jnz .pixel_loop

 ret
%endmacro

M7__REPEAT 0, 0, M7_REPEAT, M7_REPEAT_Y
M7__REPEAT 1, 0, M7_REPEAT_X, M7_REPEAT_XY
M7__CHAR0  0, 0, M7_CHAR0, M7_CHAR0_Y
M7__CHAR0  1, 0, M7_CHAR0_X, M7_CHAR0_XY
M7__CLIP   0, 0, M7_CLIP, M7_CLIP_Y
M7__CLIP   1, 0, M7_CLIP_X, M7_CLIP_XY

M7__REPEAT 0, 1, M7P_REPEAT, M7P_REPEAT_Y
M7__REPEAT 1, 1, M7P_REPEAT_X, M7P_REPEAT_XY
M7__CHAR0  0, 1, M7P_CHAR0, M7P_CHAR0_Y
M7__CHAR0  1, 1, M7P_CHAR0_X, M7P_CHAR0_XY
M7__CLIP   0, 1, M7P_CLIP, M7P_CLIP_Y
M7__CLIP   1, 1, M7P_CLIP_X, M7P_CLIP_XY

EXPORT SCREEN_MODE_7    ; There is 1 version of the render code!
 test bl,0x10   ; Are sprites enabled?
 jz .no_sprites

 mov al,[Redo_OAM]
 test al,al
 jz .ready_oam

 push ebx
 pusha
 call Recache_OAM
 popa
 pop ebx

.ready_oam:
 and ebx,byte 3
 jz near NO_PLANES
 and ebx,byte 1
 jnz M7_P1
 ; we only reach here when bit 0 clear, bit 1 set
 test byte [_SETINI],0x40
 jnz near M7P_P1

 ret

ALIGNC
.no_sprites:
 and ebx,byte 3
 jz .no_layer
 and ebx,byte 1
 jnz near RENDER_MODE_7
 ; we only reach here when bit 0 clear, bit 1 set
 test byte [_SETINI],0x40
 jnz near RENDER_MODE_7

.no_layer:
 ret

ALIGNC
M7_P1:
 ALL_SPRITES_PLANE 3
 ALL_SPRITES_PLANE 2
 ALL_SPRITES_PLANE 1
 call RENDER_MODE_7
 ALL_SPRITES_PLANE 0
 ret

ALIGNC
M7P_P1:
 ALL_SPRITES_PLANE 3
 ALL_SPRITES_PLANE 2
 ALL_SPRITES_PLANE 1
 mov byte [Tile_priority_opcode],0x79   ;jns
 call RENDER_MODE_7
 ALL_SPRITES_PLANE 0
 mov byte [Tile_priority_opcode],0x78   ;js
 call RENDER_MODE_7
 ret

ALIGNC
EXPORT_C Reset_Mode_7
 ; Set eax to 0, as we're setting most everything to 0...
 xor eax,eax

 mov [_M7SEL],al
 mov byte [Redo_M7],0x0F
 mov dword [M7_Handler],M7_REPEAT
 mov byte [Redo_16x8],0
 mov [MPYL],eax
 mov [_M7A],eax
 mov [_M7B],eax
 mov [_M7C],eax
 mov [_M7D],eax
 mov [_M7X_13],eax
 mov [_M7Y_13],eax
 mov [_M7H],eax
 mov [_M7V],eax
 mov [_M7X],eax
 mov [_M7Y],eax

 ret

Do_16x8_Multiply:
 push ebx
 mov ebx,edi
 movsx edi,byte [_M7B+1]
 mov byte [Redo_16x8],0
 imul edi,[_M7A]    ; I think signed is used makes most sense!
 mov [MPYL],edi
 mov al,[ebx]
 pop ebx
 ret

ALIGNC
EXPORT SNES_R2134 ; MPYL
 mov edi,MPYL
 cmp byte [Redo_16x8],0
 jnz Do_16x8_Multiply
 mov al,[edi]
 ret

ALIGNC
EXPORT SNES_R2135 ; MPYM
 mov edi,MPYM
 cmp byte [Redo_16x8],0
 jnz Do_16x8_Multiply
 mov al,[edi]
 ret

ALIGNC
EXPORT SNES_R2136 ; MPYH
 mov edi,MPYH
 cmp byte [Redo_16x8],0
 jnz Do_16x8_Multiply
 mov al,[edi]
 ret

ALIGNC
EXPORT SNES_W211A ; M7SEL   ; New for 0.12
 cmp al,[_M7SEL]
 je .no_change
 SyncRender     ;*
 ; Note: M7_Table,M7_Handler are declared in ScreenL.S
 push ebx
 mov bl,al
 mov [_M7SEL],al
 shr ebx,4
 and al,3
 and ebx,0x0C
 or bl,al

 mov al,[_SETINI]   ; 0.42 - Mode 7 w/ priority
 and al,0x40
 shr al,2
 or bl,al

 mov ebx,[M7_Table+ebx*4]
 mov [M7_Handler],ebx
 pop ebx
 mov al,[_M7SEL]

.no_change:
 ret

ALIGNC
EXPORT SNES_W211B ; M7A
 SyncRender     ;*M7
 ; Used for matrix render and 16-bit M7A * 8-bit = 24-bit multiply
 push eax
 mov ah,al
 mov al,[_M7A+1]
 cwde
 mov [_M7A],eax
 mov al,0x01    ; Recalculate A
 or [Redo_M7],al
 mov byte [Redo_16x8],-1
 pop eax
 ret

ALIGNC
EXPORT SNES_W211C ; M7B
 SyncRender     ;*M7
 ; Used for matrix render and 16-bit * 8-bit M7B high byte = 24-bit multiply
 push eax
 mov ah,al
 mov al,[_M7B+1]
 cwde
 mov [_M7B],eax
 mov al,0x02    ; Recalculate B
 or [Redo_M7],al
 mov byte [Redo_16x8],-1
 pop eax
 ret

ALIGNC
EXPORT SNES_W211D ; M7C
 SyncRender     ;*M7
 push eax
 mov ah,al
 mov al,[_M7C+1]
 cwde
 mov [_M7C],eax
 mov al,0x04    ; Recalculate C
 or [Redo_M7],al
 pop eax
 ret

ALIGNC
EXPORT SNES_W211E ; M7D
 SyncRender     ;*M7
 push eax
 mov ah,al
 mov al,[_M7D+1]
 cwde
 mov [_M7D],eax
 mov al,0x08    ; Recalculate D
 or [Redo_M7],al
 pop eax
 ret

ALIGNC
EXPORT SNES_W211F ; M7X
 SyncRender     ;*M7
 push eax
 mov ah,al
 mov al,[_M7X+1]
 mov [_M7X],ax

 shl eax,0x13
 sar eax,0x13
 mov [_M7X_13],eax
 mov al,0x05    ; Recalculate A,C
 or [Redo_M7],al
 pop eax
 ret

ALIGNC
EXPORT SNES_W2120 ; M7Y
 SyncRender     ;*M7
 push eax
 mov ah,al
 mov al,[_M7Y+1]
 mov [_M7Y],ax

 shl eax,0x13
 sar eax,0x13
 mov [_M7Y_13],eax
 mov al,0x0E    ; Recalculate B,C,D
 or [Redo_M7],al
 pop eax
 ret

section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
