#include "wrapaleg.h"
#include "helper.h"

BEGIN_DIGI_DRIVER_LIST
 DIGI_DRIVER_SOUNDSCAPE
 DIGI_DRIVER_AUDIODRIVE
 DIGI_DRIVER_WINSOUNDSYS
 DIGI_DRIVER_SB
END_DIGI_DRIVER_LIST

BEGIN_MIDI_DRIVER_LIST
END_MIDI_DRIVER_LIST

// TRAC 23/07/1999: Added some code to allow ADSR to be disabled at runtime

#define FREQ 32000      //32kHz true SNES mix speed
#define LAG  10         //Lag between sound mixed and sound heard (1/LAG sec)

#define ADSR
//#define ADSR_OPTIONAL

#define ATTACK 0    //A of ADSR
#define DECAY 1     //D of ADSR
#define SUSTAIN 2   //S of ADSR
#define RELEASE 3   //R of ADSR
#define DECREASE 4  //GAIN linear decrease mode
#define EXP 5       //GAIN exponential decrease mode
#define INCREASE 6  //GAIN linear increase mode
#define BENT 7      //GAIN bent line increase mode
#define DIRECT 8    //Directly specify ENVX

extern unsigned char SPCRAM[65536];
extern unsigned long SPC_Cycles;

unsigned long SPC_TIMER0,SPC_TIMER1,SPC_TIMER2;
unsigned long TotalCycles,CycleLatch0,CycleLatch1,CycleLatch2;
unsigned long SPC_TIMT0 ,SPC_TIMT1 ,SPC_TIMT2;
unsigned char SPC_CNT0  ,SPC_CNT1  ,SPC_CNT2  ,
          SPC_CTRL  ,SPC_MASK,
          SPC_DSP_ADDR, SPC_DSP_DATA;
unsigned long STRlatch;

static SAMPLE *STRsamp;

#ifdef ADSR_OPTIONAL
unsigned char ADSR_ENABLED = -1;
#endif
unsigned char SOUND_ENABLED;

unsigned char ENVX_ENABLED = -1;

unsigned char SPC_DSP[256];

// INTERNAL (static) STUFF

static struct {
    short buf[16];
    unsigned int envx,envcyc;
    int envstate;
    int ar,dr,sl,sr,gn;
    unsigned long bufptr;
    unsigned int brrptr;
    unsigned short *loopptr;
    short last1,last2;
} SNDvoices[8];

static int SNDkeys;
static const unsigned int C[0x20]={0x0,0x20000,0x18000,0x14000,0x10000,0xC000,
 0xA000,0x8000,0x6000,0x5000,0x4000,0x3000,0x2800,0x2000,0x1800,0x1400,0x1000,
 0xC00,0xA00,0x800,0x600,0x500,0x400,0x300,0x280,0x200,0x180,0x140,0x100,0xC0,
 0x80,0x40};    //How many cycles till adjust ADSR/GAIN

// OLD STUFF

void Update_SPC_Timer_0(){
 if (!(SPC_CTRL & 1)) return;
 SPC_TIMER0 += (TotalCycles - CycleLatch0) / 256;
 CycleLatch0 += (TotalCycles - CycleLatch0) & ~0xFF;

 SPC_CNT0 = (SPC_CNT0 + (SPC_TIMER0 / SPC_TIMT0)) & 0xF;
 SPC_TIMER0 %= SPC_TIMT0;
}

void Update_SPC_Timer_1(){
 if (!(SPC_CTRL & 2)) return;
 SPC_TIMER1 += (TotalCycles - CycleLatch1) / 256;
 CycleLatch1 += (TotalCycles - CycleLatch1) & ~0xFF;

 SPC_CNT1 = (SPC_CNT1 + (SPC_TIMER1 / SPC_TIMT1)) & 0xF;
 SPC_TIMER1 %= SPC_TIMT1;
}

void Update_SPC_Timer_2(){
 if (!(SPC_CTRL & 4)) return;
 SPC_TIMER2 += (TotalCycles - CycleLatch2) / 32;
 CycleLatch2 += (TotalCycles - CycleLatch2) & ~0x1F;

 SPC_CNT2 = (SPC_CNT2 + (SPC_TIMER2 / SPC_TIMT2)) & 0xF;
 SPC_TIMER2 %= SPC_TIMT2;
}

void Wrap_SPC_Cyclecounter(){
 int c;

 for (c = 0; c < 8; c++) SNDvoices[c].envcyc -= 0xF0000000;
 TotalCycles -= 0xF0000000;
 SPC_Cycles -= 0xF0000000;
 CycleLatch0 -= 0xF0000000;
 CycleLatch1 -= 0xF0000000;
 CycleLatch2 -= 0xF0000000;
}

// NEW PRIVATE (static) STUFF

static int get_chunk(int v)
{
    int i,output;
    unsigned int brrptr;
    short last1,last2;
    unsigned char header,range,filter,input;

    brrptr=SNDvoices[v].brrptr;
    last1=SNDvoices[v].last1;
    last2=SNDvoices[v].last2;

    if(brrptr>=0x10000) return 1;

    header=SPCRAM[brrptr++];
    filter=(header>>2)&3;
    range=header>>4;
    i=0;
    while(i<16)
    {
        if(brrptr>=0x10000){
         if (header & 1)
         {
          if (header & 2)
           SNDvoices[v].brrptr=*SNDvoices[v].loopptr;
          else
           SNDvoices[v].brrptr=0x10000;
          SPC_DSP[0x7C] |= 1<<v;
         } else
          SNDvoices[v].brrptr=brrptr;

         SNDvoices[v].last1=last1;
         SNDvoices[v].last2=last2;
         return 1;
        }

        input=SPCRAM[brrptr++];

        output=(input>>4)&0x0F;
        output=(output^8)-8;
        output<<=range;
        switch(filter)
        {
        case 1:
            output+=(int)(last1*15/16);
            break;
        case 2:
            output+=(int)((last1*61/32)-(last2*15/16));
            break;
        case 3:
            output+=(int)((last1*115/64)-(last2*13/16));
            break;
        }
        if(output>0x7FFF)
            output=0x7FFF;
        else if(output<-0x8000)
            output=-0x8000;
        last2=last1;
        last1=output;
        SNDvoices[v].buf[i++]=output;

        output=input&0x0F;
        output=(output^8)-8;
        output<<=range;
        switch(filter)
        {
        case 1:
            output+=(int)(last1*15/16);
            break;
        case 2:
            output+=(int)((last1*61/32)-(last2*15/16));
            break;
        case 3:
            output+=(int)((last1*115/64)-(last2*13/16));
            break;
        }
        if(output>0x7FFF)
            output=0x7FFF;
        else if(output<-0x8000)
            output=-0x8000;
        last2=last1;
        last1=output;
        SNDvoices[v].buf[i++]=output;
    }
    if(header&1)
    {
        if(header&2)
            SNDvoices[v].brrptr=*SNDvoices[v].loopptr;
        else
            SNDvoices[v].brrptr=0x10000;
        SPC_DSP[0x7C]|=1<<v;
    }
    else
        SNDvoices[v].brrptr=brrptr;
    SNDvoices[v].last1=last1;
    SNDvoices[v].last2=last2;
    return 0;
}

#if defined(ADSR) || defined(ADSR_OPTIONAL)
static int SNDDoEnv(int voice)
{
    unsigned int envx,cyc,c;
    envx=SNDvoices[voice].envx;
    for(;;)
    {
        cyc=TotalCycles-SNDvoices[voice].envcyc;
        switch (SNDvoices[voice].envstate)
        {
            case ATTACK:
            c=C[(SNDvoices[voice].ar<<1)+1];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx+=0x2000000;        //add 1/64th
                if(envx>=0x7F000000)
                {
                    envx=0x7F000000;
                    if(SNDvoices[voice].sl!=7) SNDvoices[voice].envstate=DECAY;
                    else SNDvoices[voice].envstate=SUSTAIN;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case DECAY:
            c=C[(SNDvoices[voice].dr<<1)+0x10];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx=(envx>>8)*255;     //mult by 1-1/256
                if(envx<=0x10000000*(SNDvoices[voice].sl+1))
                {
                    envx=0x10000000*(SNDvoices[voice].sl+1);
                    SNDvoices[voice].envstate=SUSTAIN;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case SUSTAIN:
            c=C[SNDvoices[voice].sr];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx=(envx>>8)*255;     //mult by 1-1/256
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case RELEASE:
            //says add 1/256??  That won't release, must be subtract.
            //But how often?  Oh well, who cares, I'll just
            //pick a number. :)
            c=C[0x19];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx-=0x800000;         //sub 1/256th
                if((envx==0)||(envx>0x7F000000))
                {
                    SNDkeys&=~(1<<voice);
                    return SNDvoices[voice].envx=0;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case INCREASE:
            c=C[SNDvoices[voice].gn];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx+=0x2000000;        //add 1/64th
                if(envx>0x7F000000)
                {
                    SNDvoices[voice].envcyc=TotalCycles;
                    return SNDvoices[voice].envx=0x7F000000;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case DECREASE:
            c=C[SNDvoices[voice].gn];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx-=0x2000000;        //sub 1/64th
                if(envx>0x7F000000)     //underflow
                {
                    SNDvoices[voice].envcyc=TotalCycles;
                    return SNDvoices[voice].envx=0;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case EXP:
            c=C[SNDvoices[voice].gn];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                envx=(envx>>8)*255;     //mult by 1-1/256
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case BENT:
            c=C[SNDvoices[voice].gn];
            if(c==0)
            {
                SNDvoices[voice].envcyc=TotalCycles;
                return SNDvoices[voice].envx=envx;
            }
            if(cyc>c)
            {
                SNDvoices[voice].envcyc+=c;
                if(envx<0x60000000)
                 envx+=0x2000000;//add 1/64th
                else
                 envx+=0x800000; //add 1/256th
                if(envx>0x7F000000)
                {
                    SNDvoices[voice].envcyc=TotalCycles;
                    return SNDvoices[voice].envx=0x7F000000;
                }
            }
            else return SNDvoices[voice].envx=envx;
            break;

            case DIRECT:
            SNDvoices[voice].envcyc=TotalCycles;
            return envx;
        }
    }
}
#endif

static void SNDNoteOn(int v)
{
    int i,cursamp,adsr1,adsr2,gain;
/*  int lvol,rvol=0,pitch;*/
    struct {unsigned short start,loop;} *samp_dir;
    samp_dir=(void *)&SPCRAM[(int)SPC_DSP[0x5D]<<8];
    v&=SPC_MASK;
    SNDkeys|=v;
    SPC_DSP[0x7C]&=~v;
    for(i=0;i<8;i++)
        if(v&(1<<i))
        {
            cursamp=SPC_DSP[4+(i<<4)];
            SNDvoices[i].brrptr=samp_dir[cursamp].start;
            SNDvoices[i].loopptr=&samp_dir[cursamp].loop;
            SNDvoices[i].bufptr=0x20000;
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED){
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
            adsr1=SPC_DSP[(i<<4)+5];
            if(adsr1&0x80)
            {
                //ADSR mode
                adsr2=SPC_DSP[(i<<4)+6];
                SNDvoices[i].envx=0;
                SNDvoices[i].envcyc=TotalCycles;
                SNDvoices[i].envstate=ATTACK;
                SNDvoices[i].ar=adsr1&0xF;
                SNDvoices[i].dr=adsr1>>4&7;
                SNDvoices[i].sr=adsr2&0x1f;
                SNDvoices[i].sl=adsr2>>5;
            }
            else
            {
                //GAIN mode
                gain=SPC_DSP[(i<<4)+7];
                if(gain&0x80)
                {
                    SNDvoices[i].envx=0;
                    SNDvoices[i].envcyc=TotalCycles;
                    SNDvoices[i].envstate=gain>>5;
                    SNDvoices[i].gn=gain&0x1F;
                }
                else
                {
                    SNDvoices[i].envx=(gain&0x7F)<<24;
                    SNDvoices[i].envstate=DIRECT;
                }
            }
#endif
#ifdef ADSR_OPTIONAL
            }
#endif
    }
}

static void SNDNoteOff(int v)
{
    int i;
    v&=SNDkeys;
    for(i=0;i<8;i++)
        if(v&(1<<i))
        {
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED){
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
            SNDvoices[i].envstate=RELEASE;
#endif
#ifdef ADSR_OPTIONAL
            } else {
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
            SNDkeys&=~(1<<i);
#endif
#ifdef ADSR_OPTIONAL
            }
#endif
        }
}

// NEW SHARED STUFF

static int sound_enable_mode = 0;

void Remove_Sound()
{
 if (!STRsamp)
 {
  voice_stop(0);
  deallocate_voice(0);
  destroy_sample(STRsamp);
 }
 remove_sound();
 SOUND_ENABLED = sound_enable_mode = 0;
}

int Install_Sound(int stereo)
{
 int i;
 if (sound_enable_mode) Remove_Sound();

 reserve_voices(1, 0);
 if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL))
  return SOUND_ENABLED = sound_enable_mode = 0;

 i = FREQ / LAG;
 STRsamp = create_sample(16, stereo ? TRUE : FALSE, FREQ, i);
 if (!STRsamp)
 {
  remove_sound();
  return SOUND_ENABLED = sound_enable_mode = 0;
 }

 SPC_MASK = 0xFF;
 SNDkeys = 0;
 STRlatch = 0;

 for (i = (stereo ? (i << 1) : i) - 1; i >= 0; i--)
  ((unsigned short *) STRsamp->data)[i] = 0x8000;
 allocate_voice(STRsamp);
 voice_set_playmode(0, PLAYMODE_LOOP);
 voice_set_volume(0, 255);
 voice_set_pan(0, 128);
 voice_start(0);
 return SOUND_ENABLED = sound_enable_mode = stereo + 1;
}

void Reset_Sound()
{
 int i = FREQ / LAG;

 if (!sound_enable_mode) return;
 if (!STRsamp) return;

 voice_stop(0);
 deallocate_voice(0);

 SPC_MASK = 0xFF;
 SNDkeys = 0;
 STRlatch = 0;

 for (i = (sound_enable_mode - 1 ? (i << 1) : i) - 1; i >= 0; i--)
  ((unsigned short *) STRsamp->data)[i] = 0x8000;
 allocate_voice(STRsamp);
 voice_set_playmode(0, PLAYMODE_LOOP);
 voice_set_volume(0, 255);
 voice_set_pan(0, 128);
 voice_start(0);
}

void SNDUpdate(void)
{
    unsigned long mixend,from,to,slen,step,bufptr;
    long envx;
    unsigned short *buf;
    short *sbuf;
    int i,v,lvol,rvol,done,t;
    int outx = 0;

    if (!SPC_ENABLED || !SOUND_ENABLED) return;

    mixend=voice_get_position(0);
    if(mixend==STRlatch)
        return;
    buf=(unsigned short *)STRsamp->data;
    if(SOUND_ENABLED==2)
    {
        slen=STRsamp->len<<1;
        from=STRlatch<<1;
        to=mixend<<1;
    }
    else
    {
        slen=STRsamp->len;
        from=STRlatch;
        to=mixend;
    }
    i=from;
    while(i!=to)
    {
        buf[i]=0x8000;
        if(++i==slen)
            i=0;
    }
    SNDkeys&=SPC_MASK;
    for(v=0;v<8;v++)
    {
        if(SNDkeys&(1<<v))
        {
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED)
                envx=SNDDoEnv(v)>>24;
            else
                envx=0x7F;
#else
#ifdef ADSR
            envx=SNDDoEnv(v)>>24;
#else
            envx=0x7F;
#endif
#endif
            step=((unsigned long)(*(unsigned short *)&SPC_DSP[(v<<4)+2])<<16)
                  /FREQ;
            sbuf=SNDvoices[v].buf;
            bufptr=SNDvoices[v].bufptr;
            if(SOUND_ENABLED==2)
            {
                lvol=((int)((signed char)SPC_DSP[v<<4])*
                      (int)((signed char)SPC_DSP[0x0C]))>>7;
                rvol=((int)((signed char)SPC_DSP[(v<<4)+1])*
                      (int)((signed char)SPC_DSP[0x1C]))>>7;
                done=0;
                i=from;
                while(i!=to)
                {
                    while(bufptr>=0x20000)
                    {
                        if(get_chunk(v))
                            done=1;
                        bufptr-=0x20000;
                    }
                    if(done)
                    {
                        SNDkeys&=~(1<<v);
                        break;
                    }
                    outx=(sbuf[bufptr>>13]*envx)>>7;
                    t=(int)buf[i]+((outx*lvol)>>6);
                    if(t>0xFFFF)
                        buf[i]=0xFFFF;
                    else if(t<0)
                        buf[i]=0;
                    else
                        buf[i]=t;
                    i++;
                    t=(int)buf[i]+((outx*rvol)>>6);
                    if(t>0xFFFF)
                        buf[i]=0xFFFF;
                    else if(t<0)
                        buf[i]=0;
                    else
                        buf[i]=t;
                    if(++i>=slen)
                        i=0;
                    bufptr+=step;
                }
            }
            else
            {
                lvol=(int)((signed char)SPC_DSP[v<<4])*
                     (int)((signed char)SPC_DSP[0x0C]);
                if(lvol<0)
                    lvol=-lvol;
                rvol=(int)((signed char)SPC_DSP[(v<<4)+1])*
                     (int)((signed char)SPC_DSP[0x1C]);
                if(rvol<0)
                    rvol=-rvol;
                lvol=(lvol+rvol)>>8;
                done=0;
                i=from;
                while(i!=to)
                {
                    while(bufptr>=0x20000)
                    {
                        if(get_chunk(v))
                            done=1;
                        bufptr-=0x20000;
                    }
                    if(done)
                    {
                        SNDkeys&=~(1<<v);
                        break;
                    }
                    outx=(sbuf[bufptr>>13]*envx)>>7;
                    t=(int)buf[i]+((outx*lvol)>>6);
                    if(t>0xFFFF)
                        buf[i]=0xFFFF;
                    else if(t<0)
                        buf[i]=0;
                    else
                        buf[i]=t;
                    if(++i>=slen)
                        i=0;
                    bufptr+=step;
                }
            }
            SNDvoices[v].bufptr=bufptr;
            SPC_DSP[(v<<4)+9]=outx>>8;
        }
        else
            SPC_DSP[(v<<4)+9]=0;
    }
    STRlatch=mixend;
}

void SPC_READ_DSP()
{
    int addr_lo=SPC_DSP_ADDR&0xF, addr_hi=(SPC_DSP_ADDR>>4)&7;
    SNDUpdate();
    switch(addr_lo)
    {
    case 0x8:   SPC_DSP[SPC_DSP_ADDR] =
                 ENVX_ENABLED ? SNDDoEnv(addr_hi) >> 24 : 0;
                break;
    case 0xc:   switch(addr_hi)
                {
                case 0x4:   SPC_DSP[SPC_DSP_ADDR]=SNDkeys;
                            break;
                }
    }
}

void SPC_WRITE_DSP()
{
    int i;
    int addr_lo=SPC_DSP_ADDR&0xF, addr_hi=SPC_DSP_ADDR>>4;
    SNDUpdate();
    switch(addr_lo)
    {
    //break just means nothing needs to be done right now
    //Commented cases are unsupported registers
    // These go for individual voices
    case 0x0:   //Volume left
                break;
    case 0x1:   //Volume right
                break;
    case 0x2:   //Pitch lo
                break;
    case 0x3:   //Pitch hi
                SPC_DSP_DATA&=0x3F;
                break;
    case 0x4:   //Source number
                break;
#if defined(ADSR) || defined(ADSR_OPTIONAL)
    case 0x5:   //ADSR1
#ifdef ADSR_OPTIONAL
                if(!ADSR_ENABLED) break;
#endif
                if(!SOUND_ENABLED) break;
                if((SNDkeys&(1<<addr_hi))&&
                  ((SPC_DSP_DATA&0x80)!=(SPC_DSP[SPC_DSP_ADDR]&0x80)))
                {
                    SNDDoEnv(addr_hi);
                    if(SPC_DSP_DATA&0x80)
                    {
                        //switch to ADSR--not sure what to do
                        i=SPC_DSP[(addr_hi<<4)+6];
                        SNDvoices[addr_hi].envstate=ATTACK;
                        SNDvoices[addr_hi].ar=SPC_DSP_DATA&0xF;
                        SNDvoices[addr_hi].dr=SPC_DSP_DATA>>4&7;
                        SNDvoices[addr_hi].sr=i&0x1f;
                        SNDvoices[addr_hi].sl=i>>5;
                    }
                    else
                    {
                        //switch to a GAIN mode
                        i=SPC_DSP[(addr_hi<<4)+7];
                        if(i&0x80)
                        {
                            SNDvoices[addr_hi].envstate=i>>5;
                            SNDvoices[addr_hi].gn=i&0x1F;
                        }
                        else
                        {
                            SNDvoices[addr_hi].envx=(i&0x7F)<<24;
                            SNDvoices[addr_hi].envstate=DIRECT;
                        }
                    }
                }
                break;
    case 0x6:   //ADSR2
#ifdef ADSR_OPTIONAL
                if(!ADSR_ENABLED) break;
#endif
                if(!SOUND_ENABLED) break;
                if(SNDkeys&(1<<addr_hi))
                    SNDDoEnv(addr_hi);
                SNDvoices[addr_hi].sr=SPC_DSP_DATA&0x1f;
                SNDvoices[addr_hi].sl=SPC_DSP_DATA>>5;
                break;
    case 0x7:   //GAIN
#ifdef ADSR_OPTIONAL
                if(!ADSR_ENABLED) break;
#endif
                if(!SOUND_ENABLED) break;
                if((SNDkeys&(1<<addr_hi))&&
                  (SPC_DSP_DATA!=SPC_DSP[SPC_DSP_ADDR])&&
                  !(SPC_DSP[(addr_hi<<4)+5]&0x80))
                {
                    if(SPC_DSP_DATA&0x80)
                    {
                        SNDDoEnv(addr_hi);
                        SNDvoices[addr_hi].envstate=SPC_DSP_DATA>>5;
                        SNDvoices[addr_hi].gn=SPC_DSP_DATA&0x1F;
                    }
                    else
                    {
                        SNDvoices[addr_hi].envx=(SPC_DSP_DATA&0x7F)<<24;
                        SNDvoices[addr_hi].envstate=DIRECT;
                    }
                }
                break;
#endif
    // These are general registers
    case 0xc:
                switch(addr_hi)
                {
                case 0: //Left main volume
                        break;
                case 1: //Right main volume
                        break;
//              case 2: //Left echo volume
//                      break;
//              case 3: //Right echo volume
//                      break;
                case 4: //Key on
                        if(SOUND_ENABLED)
                            SNDNoteOn(SPC_DSP_DATA);
                        break;
                case 5: //Key off
                        if(SOUND_ENABLED)
                            SNDNoteOff(SPC_DSP_DATA);
                        SPC_DSP_DATA=0;
                        break;
//              case 6: //Clocks
//                      break;
                case 7: //ENDX
                        SPC_DSP_DATA=0;
                        break;
                }
                break;
    case 0xd:
                switch(addr_hi)
                {
//              case 0: //Echo feedback
//                      break;
//              case 1: //Not a valid register
//                      break;
//              case 2: //Modulation
//                      break;
//              case 3: //Noise toggle
//                      break;
//              case 4: //Echo toggle
//                      break;
                case 5: //Source directory address
                        break;
//              case 6: //Echo start address
//                      break;
//              case 7: //Echo delay
//                      break;
                }
//  case 0xf:       //FIR Echo filter
//              break;
    }
    SPC_DSP[SPC_DSP_ADDR]=SPC_DSP_DATA;
}

void sound_pause(void)
{
    if(SOUND_ENABLED)
        voice_stop(0);
}

void sound_resume(void)
{
    if(SOUND_ENABLED)
        voice_start(0);
}
