//#define Profile_Recache_OAM
/*

 Sprite render functions

 In all the sprite renderers...
  eax = (internal)
  ebx = (internal)
  cl  = sprite counter for priority (internal)
  ch  = sprite counter for set      (internal)
  dl  = sprite priority identifier  (input)
  dh  = (internal)
  esi = current sprite              (internal)
  edi = (internal)
*/

#include "misc.i"
#include "screen.i"

EXPORT_C sprites_start

.data
ALIGND
// Sprite offset tables moved to ScreenL.S
Sprite_Mode_Table:
.long   RENDER_SPRITE_8x8_16x16     // 8x8 or 16x16
.long   RENDER_SPRITE_8x8_32x32     // 8x8 or 32x32
.long   RENDER_SPRITE_8x8_64x64     // 8x8 or 64x64
.long   RENDER_SPRITE_16x16_32x32   // 16x16 or 32x32
.long   RENDER_SPRITE_16x16_64x64   // 16x16 or 64x64
.long   RENDER_SPRITE_32x32_64x64   // 32x32 or 64x64
.long   SPRITE_NOT_SUPPORTED        // Not supported in SNES hardware?
.long   SPRITE_NOT_SUPPORTED        // Not supported in SNES hardware?

Sprite_Size_Table:
.byte   8, 16, -7,-15
.byte   8, 32, -7,-31
.byte   8, 64, -7,-63
.byte  16, 32,-15,-31
.byte  16, 64,-15,-63
.byte  32, 64,-31,-63
.byte   0,  0,  0,  0
.byte   0,  0,  0,  0

.bss
ALIGND
.globl OAM_X,OAM_Tile,OAM_Size,OAM_Y,OAM_Palette,OAM_Flip,_OAM
OAM_X:          skipl 128       // OAM sprite X positions
OAM_Tile:       skipl 128       // OAM sprite tile addresses
OAM_Size:       skipb 128       // OAM sprite size flags
OAM_Y:          skipb 128       // OAM sprite Y positions
OAM_Palette:    skipb 128       // OAM sprite palettes
OAM_Flip:       skipb 128       // OAM sprite X/Y flip
OAM_Priority:   skipb 128       // OAM sprite priorities
_OAM:           skipb 512+32    // Buffer for OAM
// Renderer for current sprite sizes
.globl SpriteCount,Sprite_Mode,_HiSprite,SPRLatch,_OBSEL,OAMOdd,OAMReadOdd
SpriteCount:skipl
Sprite_Mode:skipl
_HiSprite:  skipl
SPRLatch:   skipb   // Sprite Priority Rotation latch flag
_OBSEL:     skipb   // sssnnxbb  sss=sprite size,nn=upper 4k address,bb=offset
OAMOdd:     skipb
OAMReadOdd: skipb
.globl _HiSpriteCnt1,_HiSpriteCnt2,_OBBASE,_OBNAME
_HiSpriteCnt1:  skipl   // First set size and bit offset
_HiSpriteCnt2:  skipl   // Second set size and bit offset
_OBBASE:        skipl   // VRAM location of sprite tiles 00-FF
_OBNAME:        skipl   // VRAM location of sprite tiles 100-1FF
.globl _OAMAddress,_OAMAddress_VBL
_OAMAddress:    skipl
_OAMAddress_VBL:skipl   // Restore this at VBL
.globl _HiSpriteAddr,_HiSpriteBits
_HiSpriteAddr:  skipl   // OAM address of sprite in 512b table
_HiSpriteBits:  skipl   // OAM address of sprite in 32b table
.globl NextSprite,Priority_Used
NextSprite:     skipl
Priority_Used:  skipl
.globl SpriteZ,Rendered_Sprites,SpriteList
ALIGND
SpriteZ:    skipb 64+256+64
ALIGND
Rendered_Sprites:   skipb 64+256+64
ALIGND
SpriteList: skipb 129

.text
ALIGNC
// This covers special cases or not implemented features!
SPRITE_NOT_SUPPORTED:
  movl Priority_Used,%eax
  testl %eax,%eax
  jz  2f
  
  movl $Rendered_Sprites+64,%edi
  xorl %eax,%eax
  movl $256/32,%ebp
  movl %eax,Priority_Used

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

  movl $SpriteZ+64,%edi
  movl $256/32,%ebp
  decl %eax

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

2:ret

.macro PlotSprite Tilesize
// Don't draw unless sprite is on this scanline
  movb Current_Line_Render,%al
  decl %eax
  subb OAM_Y(%esi),%al       // Get Y position
  cmpb $\Tilesize,%al
  jae 7f
  movb OAM_Priority(%esi),%bl
  movb OAM_Flip(%esi),%dl
  andl $3,%ebx
  andl $0x3F,%eax
  movb %bl,%cl
  movb OAM_Palette(%esi),%dh
  movb $-1,Priority_Used(%ebx)
  movl OAM_X(,%esi,4),%ebp   // X position
  movl $Rendered_Sprites+64,%edi
//movl ScreenDestPtr,%edi
  movl $SpriteZ+64,%ebx
  addl %ebp,%edi             // X position in render-buffer
  addl %ebp,%ebx             // X position in Z-buffer
  movl %edx,%ebp
  movl OAM_Tile(,%esi,4),%esi
//movb %dh,ColourBase

  // We have offset into sprite offset table
  movl Tile_Offset_Table_64_8(,%eax,4),%eax
  testl $8,%ebp
  jz  1f

  xorl $-1,%eax
.if \Tilesize-8  // Not 8x8
.if \Tilesize-16 // Not 16x16
.if \Tilesize-32 // Not 32x32
  addl $1+(7*16*64+7*8),%eax
.else
  addl $1+(3*16*64+7*8),%eax
.endif
.else
  addl $1+(16*64+7*8),%eax
.endif
.else
  addl $1+(7*8),%eax
.endif

1:addl %eax,%esi
  andl $4,%ebp       // ebp = X

.if \Tilesize-8  // Not 8-bit
.if \Tilesize-16 // Not 16-bit
.if \Tilesize-32 // Not 32-bit
  call *Sprite_Tile_Plot_64_4(%ebp)
.else
  call *Sprite_Tile_Plot_32_4(%ebp)
.endif
.else
  call *Sprite_Tile_Plot_16_4(%ebp)
.endif
.else
  call *Sprite_Tile_Plot_8_4(%ebp)
.endif

7:decb %ch
  jnz 0b            // Loop for all sprites in set
  ret
.endm

.macro RenderSprite small,large,name
ALIGNC
.globl RENDER_SPRITE_\name
RENDER_SPRITE_\name:
  movb %al,%ch

  movl Priority_Used,%eax
  testl %eax,%eax
  jz  2f
  
  movl $Rendered_Sprites+64,%edi
  xorl %eax,%eax
  movl $256/32,%ebp
  movl %eax,Priority_Used

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

  movl $SpriteZ+64,%edi
  movl $256/32,%ebp
  decl %eax

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

ALIGNC
2:
0:movl NextSprite,%esi
  movl (%esi),%esi
  andl $0xFF,%esi
  incl NextSprite
  movb OAM_Size(%esi),%al
  testb %al,%al
  jnz 2f
  PlotSprite \small

ALIGNC
2:PlotSprite \large

.endm

RenderSprite 8,16,8x8_16x16
RenderSprite 8,32,8x8_32x32
RenderSprite 8,64,8x8_64x64
RenderSprite 16,32,16x16_32x32
RenderSprite 16,64,16x16_64x64
RenderSprite 32,64,32x32_64x64

.macro Plot_Sprites_2 Offset
  movb \Offset(%ebx),%al
  movb \Offset+4(%ebx),%dl
  cmpb %al,%cl
  jne 1f
  movb \Offset(%edi),%al
  testb %al,%al
  jnz 1f
  movb \Offset(%esi),%al
  movb %al,\Offset(%edi)
1:
  cmpb %dl,%cl
  jne 1f
  movb \Offset+4(%edi),%dl
  testb %dl,%dl
  jnz 1f
  movb \Offset+4(%esi),%dl
  movb %dl,\Offset+4(%edi)
1:
.endm

.globl Plot_Sprites
ALIGNC
Plot_Sprites:
  movl $SpriteZ+64,%ebx
  movl $Rendered_Sprites+64,%esi
  movl ScreenDestPtr,%edi
  movl $256/8,%ebp
0:Plot_Sprites_2 Offset=0
  Plot_Sprites_2 Offset=1
  Plot_Sprites_2 Offset=2
  Plot_Sprites_2 Offset=3

  addl $8,%ebx
  addl $8,%esi
  addl $8,%edi
  decl %ebp
  jnz 0b
  ret

/*
 In the precalculator...
  eax = (internal)
  edx = (internal)
  ebx = current sprite          (internal)
  cl  = sprite flags bit count  (internal)
  ebp = (internal)
  esi = sprite tables           (internal)
  edi = sprite flags            (internal)
*/

ALIGNC
EXPORT_C Recache_OAM
EXPORT Recache_OAM
// Need to rewrite this to calculate in order of render, and
//  to generate sprite lists for each priority!
//  (as this is only called when sprites are enabled, and OBSEL, OAM, or
//   highest priority sprite has been changed, and each layer would have
//   to decipher list on its own, added overhead should be minimal when
//   called and easily made up for when not called
#ifdef Profile_Recache_OAM
  incl _Calls_Recache_OAM
#endif
  xorl %edx,%edx
  movl _HiSprite,%ebx
  movb %dl,SpriteList
  movl _HiSpriteCnt1,%ecx // Size of first set and bit offset
  movl _HiSpriteCnt2,%eax
  movl _HiSpriteAddr,%esi
  testb %ah,%ah
  movl _HiSpriteBits,%edi
  jz  0f
  pushl $0
0:pushl %eax

/*
// Tile attribute word: YXPP CCCT TTTT TTTT
//  Where:
//   Y, X are vertical/horizontal flip
//   P is priority
//   C is color palette selector
//   T is tile number
*/

ALIGNC
0:movb (%edi),%dh
  shlb %cl,%dh
  jnc 1f
  addb %dh,%dh
  sbbl %eax,%eax
  movb (%esi),%al   // eax = X
  movb $0xFF,OAM_Size(%ebx)
  cmpl $-15,%eax
.globl sprlim_large
sprlim_large=.-1
  jge 2f
  jmp 4f

ALIGNC
1:addb %dh,%dh
  sbbl %eax,%eax
  movb (%esi),%al   // eax = X
  movb $0,OAM_Size(%ebx)
  cmpl $-7,%eax
.globl sprlim_small
sprlim_small=.-1
  jl  4f
2:
  movb 3(%esi),%dl
  movl %eax,OAM_X(,%ebx,4)
  shrl $4,%edx
  andl $3,%edx
  movb %dl,OAM_Priority(%ebx)
  xorl %eax,%eax
  movb SpriteList,%al
  incl %eax
  movb %al,SpriteList
  movb %bl,SpriteList(%eax)
  movb 1(%esi),%al
  movl _OBBASE,%ebp
  movb 2(%esi),%dl
  movb %al,OAM_Y(%ebx)
  movb 3(%esi),%dh

  movb %dh,%al
  andb $1,%dh
  jz 3f
  movl _OBNAME,%ebp

3:shll $6,%edx      // tile * 64
  movb %al,%ah
  shlb $3,%al
  addl %edx,%ebp
  shrb $4,%ah
  andl $0x1FFFF,%ebp
  orb  $0x8F,%al    // ColourBase = 1CCC 1111
  movl %ebp,OAM_Tile(,%ebx,4)
  andb $0x0C,%ah    // 0000 yx00
  movb %al,OAM_Palette(%ebx)
  movb %ah,OAM_Flip(%ebx)

ALIGNC
4:subb $2,%cl
  jns 9f
  andb $7,%cl
  incl %edi         // Goto next sprite byte

9:addl $4,%esi
  incb %bl
  decb %ch
  jnz 0b
  movb 1(%esp),%ch
  subl $0x80,%ebx
  subl $0x200,%esi
  subl $0x20,%edi
  addl $4,%esp
  testb %ch,%ch
  jnz 0b

  movb $0,Redo_OAM

  movl Priority_Used,%eax
  testl %eax,%eax
  jz  9f
  movb SpriteList,%al
  testb %al,%al
  jnz 9f
  xorl %eax,%eax
  
  movl $Rendered_Sprites+64,%edi
  xorl %eax,%eax
  movl $256/32,%ebp
  movl %eax,Priority_Used

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

  movl $SpriteZ+64,%edi
  movl $256/32,%ebp
  decl %eax

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

9:ret

ALIGNC
.globl _Reset_Sprites
_Reset_Sprites:
  pusha
  // Set eax to 0, as we're setting most everything to 0...
  xorl %eax,%eax

  // Reset sprite renderer vars
  movl $0x00,_HiSprite
  movl $_OAM+0x000,_HiSpriteAddr
  movl $_OAM+0x200,_HiSpriteBits
  movl $0x8007,_HiSpriteCnt1
  movl $0x0007,_HiSpriteCnt2
  movl $RENDER_SPRITE_8x8_16x16,Sprite_Mode
  movb $-1,Redo_OAM
  movb $-7,sprlim_small
  movb $-15,sprlim_large
  movl %eax,_OBBASE
  movl %eax,_OBNAME

  // Reset sprite port vars
  movl %eax,_OAMAddress
  movl %eax,_OAMAddress_VBL
  movb %al,OAMOdd
  movb %al,OAMReadOdd
  movb %al,SPRLatch
  movb %al,_OBSEL

//xorl %eax,%eax
  movl $Rendered_Sprites+64,%edi
  movl $256/32,%ebp
  movl %eax,Priority_Used

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

  movl $SpriteZ+64,%edi
  movl $256/32,%ebp
  decl %eax

ALIGNC
0:leal 16(%edi),%esi
  movl %eax,(%edi)
  movl %eax,4(%edi)
  movl %eax,8(%edi)
  movl %eax,12(%edi)
  movl %eax,(%esi)
  leal 16(%esi),%edi
  movl %eax,4(%esi)
  movl %eax,8(%esi)
  movl %eax,12(%esi)
  decl %ebp
  jnz 0b

  popa
  ret

.globl SNES_R2138,SNES_W2101,SNES_W2102,SNES_W2103,SNES_W2104
ALIGNC
SNES_R2138: // OAMDATAREAD
  pushl %ebx
  xorl %ebx,%ebx
  movb OAMReadOdd,%bl
  movl _OAMAddress,%edi
  movb _OAM(%ebx,%edi,2),%al
  xorb $1,%bl
  movb %bl,OAMReadOdd
  jnz 0f
  incl %edi
  cmpl $0x0110,%edi       // Ensure address wraps correctly!
  je  1f
  movl %edi,_OAMAddress
0:popl %ebx
  ret
ALIGNC
1:movl $0,_OAMAddress
  popl %ebx
  ret

ALIGNC
SNES_W2101: // OBSEL
  cmpb %al,_OBSEL
  je 0f
  SyncRender    //*
  pushl %ebx
  movb %al,_OBSEL         /* Get our copy of this */
  movl %eax,%ebx
  shrl $5,%eax
  andl $7,%eax
  pushl %edx
  movb Sprite_Size_Table+2(,%eax,4),%dl
  movb Sprite_Size_Table+3(,%eax,4),%dh
  movb %dl,sprlim_small
  movb %dh,sprlim_large
  popl %edx
  movl Sprite_Mode_Table(,%eax,4),%eax
  movl %eax,Sprite_Mode
  movl %ebx,%eax
  movl %eax,%edi
  andl $3<<3,%ebx         // Name address 0000 0000 000n n000
  andl $3,%edi            // Base address 0000 0000 0000 0xbb
  shll $10,%ebx           // Name is either 0x0000,0x1000,0x2000,0x3000
  shll $14,%edi           // Base is either 0x0000,0x2000,0x4000,0x6000
  addl %edi,%ebx
  andl $0xFFFF,%edi
  andl $0xFFFF,%ebx
  addl %edi,%edi
  addl %ebx,%ebx
  movl %edi,_OBBASE
  movl %ebx,_OBNAME
  popl %ebx
  movb $-1,Redo_OAM
0:ret 

ALIGNC
SNES_W2102: // OAMADDL
  SyncRender    //*
  pushl %ebx
  xorl %ebx,%ebx
  movb %bh,OAMOdd
  movb %al,%bl
  movb %bh,OAMReadOdd
  movl %ebx,_OAMAddress
  movl %ebx,_OAMAddress_VBL
  cmpb $0,SPRLatch
  jz  0f
//addl %ebx,%ebx    //*
//subl $4,%ebx
1:andl $0x7F,%ebx   //*
  cmpl %ebx,_HiSprite
  je  0f
  movl %ebx,_HiSprite
  shll $2,%ebx
  movl %ebx,_HiSpriteAddr
  shrl $4,%ebx
  addl $0x200,%ebx
  movl %ebx,_HiSpriteBits
  movl _HiSprite,%ebx
  decl %ebx //
  movb $128,%bh
  incb %bl
  subb %bl,%bh
  movb %bl,_HiSpriteCnt2+1  //
  movb %bh,_HiSpriteCnt1+1  //
  movb $7,%bh
  addb %bl,%bl
  subb %bl,%bh
  andb $7,%bh
  movb %bh,_HiSpriteCnt1    //

  movl $_OAM,%ebx
  addl %ebx,_HiSpriteAddr
  addl %ebx,_HiSpriteBits
0:popl %ebx
  ret

ALIGNC
SNES_W2103: // OAMADDH
  SyncRender    //*
  pushl %ebx
  xorl %ebx,%ebx
  testb %al,%al           // Is priority rotation bit set?
  movb %bl,OAMOdd
  movb $1,%bh
  movb %bl,OAMReadOdd
  jns 0f
  movb $-1,%bl
0:andb %al,%bh            // Only want MSB of address
  movb %bl,SPRLatch
  testb %bl,%bl
  movb _OAMAddress,%bl
  movb %bh,_OAMAddress_VBL+1
  movb %bh,_OAMAddress+1
  jnz 1b
  popl %ebx
  ret

ALIGNC
SNES_W2104: // OAMDATA
  pushl %ebx
  xorl %ebx,%ebx
  movb OAMOdd,%bl
  movl _OAMAddress,%edi
  cmpl $0x110,%edi
  jae 2f
  leal (%ebx,%edi,2),%edi
  cmpb %al,_OAM(%edi)
  je  0f
  pushl %edi
  SyncRender    //*
  popl %edi
  movb $-1,Redo_OAM
  movb %al,_OAM(%edi)
0:xorb $1,%bl
  movb %bl,OAMOdd
  jnz 2f
  movl _OAMAddress,%edi
  incl %edi
  cmpl $0x0110,%edi       // Ensure address wraps correctly!
  je  1f
  movl %edi,_OAMAddress
2:popl %ebx
  ret
ALIGNC
1:movl $0,_OAMAddress
  popl %ebx
  ret

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
