/*

 Tile Functions - In Assembler cos everything else is!
  Line-from-tile plotters with tile caching

 VRAM tile address in %esi + LineAddress(Y) + VRAMAddress...
 Cache tile address varies by depth...

 Tile recache
  esi = cache address   (input)
      = tile address    (internal)
  edi = cache address   (internal)
  eax,ebx,ecx,edx,ebp destroyed
 In all the plotters...
  esi = cache address   (input)
  edi = cache address   (internal)
      = screen address  (input)
 In check-first plotters...
  dl  = palette bitmask (internal)
        (except 8-bpl)
  eax = pixels          (internal)
  eax,dl destroyed
 In blank tile plotters
  Nothing
 In no-check plotters...
  ah  = palette bitmask (internal)
        (except 8-bpl)
  al  = pixels          (internal)
  dl  = pixels          (internal)
  ax,dl destroyed
*/

#include "misc.i"

EXPORT_C tiles_start

.data
ALIGND
.globl Tile_Line_8_2,Tile_Line_16_2
.globl Tile_Line_16_2_Even
.globl Tile_Line_8_4,Sprite_Tile_Plot_8_4
.globl Tile_Line_16_4,Sprite_Tile_Plot_16_4
.globl Sprite_Tile_Plot_32_4,Sprite_Tile_Plot_64_4
.globl Tile_Line_16_4_Even

Tile_Line_8_2:
Tile_Line_8_4:
    .long   PLOT8_4BplTile
    .long   PLOT8_4BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_8
    .long   Plot_Masked_X_8

Tile_Line_16_2_Even:
Tile_Line_16_4_Even:
    .long   PLOT16_4BplTile_Even
    .long   PLOT16_4BplTile_Even_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_Even_16
    .long   Plot_Masked_Even_X_16

Tile_Line_16_2:
Tile_Line_16_4:
    .long   PLOT16_4BplTile
    .long   PLOT16_4BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_16
    .long   Plot_Masked_X_16

Sprite_Tile_Plot_8_4:
    .long   PLOT8_4BplTile_Sprite
    .long   PLOT8_4BplTile_Sprite_X

Sprite_Tile_Plot_16_4:
    .long   PLOT16_4BplSprite
    .long   PLOT16_4BplSprite_X

Sprite_Tile_Plot_32_4:
    .long   PLOT32_4BplSprite
    .long   PLOT32_4BplSprite_X

Sprite_Tile_Plot_64_4:
    .long   PLOT64_4BplSprite
    .long   PLOT64_4BplSprite_X

.globl Tile_Line_8_8,Tile_Line_16_8
Tile_Line_8_8:
    .long   PLOT8_8BplTile
    .long   PLOT8_8BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_8
    .long   Plot_X_8

Tile_Line_16_8:
    .long   PLOT16_8BplTile
    .long   PLOT16_8BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_16
    .long   Plot_X_16

ALIGND
BPL0_2:
    .long   0x00000000,0xFD000000,0x00FD0000,0xFDFD0000
    .long   0x0000FD00,0xFD00FD00,0x00FDFD00,0xFDFDFD00
    .long   0x000000FD,0xFD0000FD,0x00FD00FD,0xFDFD00FD
    .long   0x0000FDFD,0xFD00FDFD,0x00FDFDFD,0xFDFDFDFD

BPL1_2:
    .long   0x00000000,0xFE000000,0x00FE0000,0xFEFE0000
    .long   0x0000FE00,0xFE00FE00,0x00FEFE00,0xFEFEFE00
    .long   0x000000FE,0xFE0000FE,0x00FE00FE,0xFEFE00FE
    .long   0x0000FEFE,0xFE00FEFE,0x00FEFEFE,0xFEFEFEFE

BPL0_4:
    .long   0x00000000,0xF1000000,0x00F10000,0xF1F10000
    .long   0x0000F100,0xF100F100,0x00F1F100,0xF1F1F100
    .long   0x000000F1,0xF10000F1,0x00F100F1,0xF1F100F1
    .long   0x0000F1F1,0xF100F1F1,0x00F1F1F1,0xF1F1F1F1

BPL1_4:
    .long   0x00000000,0xF2000000,0x00F20000,0xF2F20000
    .long   0x0000F200,0xF200F200,0x00F2F200,0xF2F2F200
    .long   0x000000F2,0xF20000F2,0x00F200F2,0xF2F200F2
    .long   0x0000F2F2,0xF200F2F2,0x00F2F2F2,0xF2F2F2F2

BPL2_4:
    .long   0x00000000,0xF4000000,0x00F40000,0xF4F40000
    .long   0x0000F400,0xF400F400,0x00F4F400,0xF4F4F400
    .long   0x000000F4,0xF40000F4,0x00F400F4,0xF4F400F4
    .long   0x0000F4F4,0xF400F4F4,0x00F4F4F4,0xF4F4F4F4

BPL3_4:
    .long   0x00000000,0xF8000000,0x00F80000,0xF8F80000
    .long   0x0000F800,0xF800F800,0x00F8F800,0xF8F8F800
    .long   0x000000F8,0xF80000F8,0x00F800F8,0xF8F800F8
    .long   0x0000F8F8,0xF800F8F8,0x00F8F8F8,0xF8F8F8F8

BPL0_8:
    .long   0x00000000,0x01000000,0x00010000,0x01010000
    .long   0x00000100,0x01000100,0x00010100,0x01010100
    .long   0x00000001,0x01000001,0x00010001,0x01010001
    .long   0x00000101,0x01000101,0x00010101,0x01010101

BPL1_8:
    .long   0x00000000,0x02000000,0x00020000,0x02020000
    .long   0x00000200,0x02000200,0x00020200,0x02020200
    .long   0x00000002,0x02000002,0x00020002,0x02020002
    .long   0x00000202,0x02000202,0x00020202,0x02020202

BPL2_8:
    .long   0x00000000,0x04000000,0x00040000,0x04040000
    .long   0x00000400,0x04000400,0x00040400,0x04040400
    .long   0x00000004,0x04000004,0x00040004,0x04040004
    .long   0x00000404,0x04000404,0x00040404,0x04040404

BPL3_8:
    .long   0x00000000,0x08000000,0x00080000,0x08080000
    .long   0x00000800,0x08000800,0x00080800,0x08080800
    .long   0x00000008,0x08000008,0x00080008,0x08080008
    .long   0x00000808,0x08000808,0x00080808,0x08080808

BPL4_8:
    .long   0x00000000,0x10000000,0x00100000,0x10100000
    .long   0x00001000,0x10001000,0x00101000,0x10101000
    .long   0x00000010,0x10000010,0x00100010,0x10100010
    .long   0x00001010,0x10001010,0x00101010,0x10101010

BPL5_8:
    .long   0x00000000,0x20000000,0x00200000,0x20200000
    .long   0x00002000,0x20002000,0x00202000,0x20202000
    .long   0x00000020,0x20000020,0x00200020,0x20200020
    .long   0x00002020,0x20002020,0x00202020,0x20202020

BPL6_8:
    .long   0x00000000,0x40000000,0x00400000,0x40400000
    .long   0x00004000,0x40004000,0x00404000,0x40404000
    .long   0x00000040,0x40000040,0x00400040,0x40400040
    .long   0x00004040,0x40004040,0x00404040,0x40404040

BPL7_8:
    .long   0x00000000,0x80000000,0x00800000,0x80800000
    .long   0x00008000,0x80008000,0x00808000,0x80808000
    .long   0x00000080,0x80000080,0x00800080,0x80800080
    .long   0x00008080,0x80008080,0x00808080,0x80808080

.bss
ALIGND

EXPORT_C TileCache2
skipl
EXPORT_C TileCache4
skipl
EXPORT_C TileCache8
skipl

.globl ColourBase,PixMask
ColourBase: skipb
PixMask:    skipb

.text
ALIGNC
.globl No_Plot
No_Plot:
    ret

.macro Plot_Masked_2 Offset,Width,Xflip,Mask=%dh
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,\Offset+1(%edi)
    movb %al,\Offset(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,(\Width-2-\Offset)(%edi)
    movb %al,(\Width-1-\Offset)(%edi)
.endif
.endm

.macro Plot_Masked_Even_2 Offset,Width,Xflip,Mask=%dh
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+2(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,\Offset+2(%edi)
    movb %al,\Offset(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+2(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,(\Width-3-\Offset)(%edi)
    movb %al,(\Width-1-\Offset)(%edi)
.endif
.endm

.macro Plot_2 Offset,Width,Xflip
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    movb %al,\Offset(%edi)
    movb %dl,\Offset+1(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    movb %al,(\Width-1-\Offset)(%edi)
    movb %dl,(\Width-2-\Offset)(%edi)
.endif
.endm

ALIGNC
.globl Plot_Masked_8
Plot_Masked_8:
    Plot_Masked_2 Offset=0 Width=8 Xflip=0
    Plot_Masked_2 Offset=2 Width=8 Xflip=0
    Plot_Masked_2 Offset=4 Width=8 Xflip=0
    Plot_Masked_2 Offset=6 Width=8 Xflip=0
    ret

ALIGNC
.globl Plot_Masked_X_8
Plot_Masked_X_8:
    Plot_Masked_2 Offset=0 Width=8 Xflip=1
    Plot_Masked_2 Offset=2 Width=8 Xflip=1
    Plot_Masked_2 Offset=4 Width=8 Xflip=1
    Plot_Masked_2 Offset=6 Width=8 Xflip=1
    ret

ALIGNC
.globl Plot_Masked_Even_16
Plot_Masked_Even_16:
//  movb ColourBase,%dh
    Plot_Masked_Even_2 Offset=0  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=4  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=8  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=12 Width=16 Xflip=0
    ret

ALIGNC
.globl Plot_Masked_Even_X_16
Plot_Masked_Even_X_16:
//  movb ColourBase,%dh
    Plot_Masked_Even_2 Offset=0  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=4  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=8  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=12 Width=16 Xflip=1
    ret

ALIGNC
.globl Plot_Masked_16
Plot_Masked_16:
//  movb ColourBase,%dh
    Plot_Masked_2 Offset=0  Width=16 Xflip=0
    Plot_Masked_2 Offset=2  Width=16 Xflip=0
    Plot_Masked_2 Offset=4  Width=16 Xflip=0
    Plot_Masked_2 Offset=6  Width=16 Xflip=0
    Plot_Masked_2 Offset=8  Width=16 Xflip=0
    Plot_Masked_2 Offset=10 Width=16 Xflip=0
    Plot_Masked_2 Offset=12 Width=16 Xflip=0
    Plot_Masked_2 Offset=14 Width=16 Xflip=0
    ret

ALIGNC
.globl Plot_Masked_X_16
Plot_Masked_X_16:
//  movb ColourBase,%dh
    Plot_Masked_2 Offset=0  Width=16 Xflip=1
    Plot_Masked_2 Offset=2  Width=16 Xflip=1
    Plot_Masked_2 Offset=4  Width=16 Xflip=1
    Plot_Masked_2 Offset=6  Width=16 Xflip=1
    Plot_Masked_2 Offset=8  Width=16 Xflip=1
    Plot_Masked_2 Offset=10 Width=16 Xflip=1
    Plot_Masked_2 Offset=12 Width=16 Xflip=1
    Plot_Masked_2 Offset=14 Width=16 Xflip=1
    ret

ALIGNC
.globl Plot_8
Plot_8:
    Plot_2 Offset=0 Width=8 Xflip=0
    Plot_2 Offset=2 Width=8 Xflip=0
    Plot_2 Offset=4 Width=8 Xflip=0
    Plot_2 Offset=6 Width=8 Xflip=0
    ret

ALIGNC
.globl Plot_X_8
Plot_X_8:
    Plot_2 Offset=0 Width=8 Xflip=1
    Plot_2 Offset=2 Width=8 Xflip=1
    Plot_2 Offset=4 Width=8 Xflip=1
    Plot_2 Offset=6 Width=8 Xflip=1
    ret

ALIGNC
.globl Plot_16
Plot_16:
    Plot_2 Offset=0  Width=16 Xflip=0
    Plot_2 Offset=2  Width=16 Xflip=0
    Plot_2 Offset=4  Width=16 Xflip=0
    Plot_2 Offset=6  Width=16 Xflip=0
    Plot_2 Offset=8  Width=16 Xflip=0
    Plot_2 Offset=10 Width=16 Xflip=0
    Plot_2 Offset=12 Width=16 Xflip=0
    Plot_2 Offset=14 Width=16 Xflip=0
    ret

ALIGNC
.globl Plot_X_16
Plot_X_16:
    Plot_2 Offset=0  Width=16 Xflip=1
    Plot_2 Offset=2  Width=16 Xflip=1
    Plot_2 Offset=4  Width=16 Xflip=1
    Plot_2 Offset=6  Width=16 Xflip=1
    Plot_2 Offset=8  Width=16 Xflip=1
    Plot_2 Offset=10 Width=16 Xflip=1
    Plot_2 Offset=12 Width=16 Xflip=1
    Plot_2 Offset=14 Width=16 Xflip=1
    ret

ALIGNC
.globl Recache_Tile_Set
Recache_Tile_Set:
    pushl %eax
    pushl %ebx
    pushl %ecx
    pushl %edx
    pushl %ebp
    pushl %esi

    movl Tile_Recache_Set_Begin,%edx
    shll $4,%edx    // address in VRAM of first 2bpl tile to recache
    xorl %ebx,%ebx
    xorl %ecx,%ecx
    movl %edx,%esi
    shll $2,%edx
    addl $_VRAM,%esi
    shll $3,%edi
    addl _TileCache2,%edx

0:  //  Bp0=*(LineAddress+0)
    movb $0x0F,%bl
    movb (%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    movl BPL0_2(,%ebx,4),%ebp
    movb $0x0F,%bl
    movl BPL0_2(%ecx),%eax

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL1_2(,%ebx,4),%ebp
    orl  BPL1_2(%ecx),%eax

    movl %eax,(%edx)
    addl $2,%esi
    movl %ebp,4(%edx)
    addl $8,%edx
    decl %edi
    jnz 0b

    movl Tile_Recache_Set_Begin,%edx
    movl Tile_Recache_Set_End,%edi
    shrl $1,%edx    // %edx = Tile_Recache_Set_Begin / 2 (rounded down)
    incl %edi
    shrl $1,%edi
    pushl %edx
    adcl $0,%edi    // %edi = (Tile_Recache_Set_End + 1) / 2 (rounded up)
    pushl %edi
    subl %edx,%edi  // Count of 4bpl tiles to recache

    shll $5,%edx    // address in VRAM of first 4bpl tile to recache
    pushl %edi
    movl %edx,%esi
    addl %edx,%edx
    addl $_VRAM,%esi
    addl _TileCache4,%edx
1:  movl $8,%edi
0:  //  Bp0=*(LineAddress+0)
    movb $0x0F,%bl
    movb (%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    movl BPL0_4(,%ebx,4),%ebp
    movb $0x0F,%bl
    movl BPL0_4(%ecx),%eax

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL1_4(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL1_4(%ecx),%eax

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL2_4(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL2_4(%ecx),%eax

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL3_4(,%ebx,4),%ebp
    orl  BPL3_4(%ecx),%eax

    movl %eax,(%edx)
    addl $2,%esi
    movl %ebp,4(%edx)
    addl $8,%edx
    decl %edi
    jnz 0b
    addl $16,%esi
    decl (%esp)
    jnz 1b

    popl %eax
    popl %edi
    popl %edx

    shrl $1,%edx    // %edx = Tile_Recache_Set_Begin / 2 (rounded down)
    incl %edi
    shrl $1,%edi
    adcl $0,%edi    // %edi = (Tile_Recache_Set_End + 1) / 2 (rounded up)
    subl %edx,%edi  // Count of 8bpl tiles to recache

    movl $_VRAM,%esi
    shll $6,%edx    // address in VRAM of first 8bpl tile to recache
    pushl %edi
    addl %edx,%esi
    addl _TileCache8,%edx
1:  movl $8,%edi
0:  //  Bp0=*(LineAddress+0)
    movb $0x0F,%bl
    movb (%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    movl BPL0_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    movl BPL0_8(%ecx),%eax

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL1_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL1_8(%ecx),%eax

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL2_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL2_8(%ecx),%eax

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL3_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL3_8(%ecx),%eax

    //  Bp4=*(LineAddress+32)
    movb 32(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL4_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL4_8(%ecx),%eax

    //  Bp5=*(LineAddress+33)
    movb 33(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL5_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL5_8(%ecx),%eax

    //  Bp6=*(LineAddress+48)
    movb 48(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL6_8(,%ebx,4),%ebp
    movb $0x0F,%bl
    orl  BPL6_8(%ecx),%eax

    //  Bp7=*(LineAddress+49)
    movb 49(%esi),%cl
    andb %cl,%bl
    andb $0xF0,%cl
    shrb $2,%cl
    orl  BPL7_8(,%ebx,4),%ebp
    orl  BPL7_8(%ecx),%eax

    movl %eax,(%edx)
    addl $2,%esi
    movl %ebp,4(%edx)
    addl $8,%edx
    decl %edi
    jnz 0b
    addl $48,%esi
    decl (%esp)
    jnz 1b

    popl %eax
    popl %esi
    popl %ebp
    popl %edx
    popl %ecx
    popl %ebx
    popl %eax
    ret

.macro P8_Plot_8 Xflip,Add=0
.if \Xflip-1
    movb \Add(%edi),%al
    movb \Add+4(%edi),%dl
    testb %al,%al
    jnz 1f
    movb (%esi),%al
    movb %al,\Add(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 4(%esi),%dl
    movb %dl,\Add+4(%edi)
1:  movb \Add+1(%edi),%al
    movb \Add+5(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 1(%esi),%al
    movb %al,\Add+1(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 5(%esi),%dl
    movb %dl,\Add+5(%edi)
1:  movb \Add+2(%edi),%al
    movb \Add+6(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 2(%esi),%al
    movb %al,\Add+2(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 6(%esi),%dl
    movb %dl,\Add+6(%edi)
1:  movb \Add+3(%edi),%al
    movb \Add+7(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 3(%esi),%al
    movb %al,\Add+3(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 7(%esi),%dl
    movb %dl,\Add+7(%edi)
1:
.else
    movb \Add(%edi),%al
    movb \Add+4(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 7(%esi),%al
    movb %al,\Add(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 3(%esi),%dl
    movb %dl,\Add+4(%edi)
1:  movb \Add+1(%edi),%al
    movb \Add+5(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 6(%esi),%al
    movb %al,\Add+1(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 2(%esi),%dl
    movb %dl,\Add+5(%edi)
1:  movb \Add+2(%edi),%al
    movb \Add+6(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 5(%esi),%al
    movb %al,\Add+2(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 1(%esi),%dl
    movb %dl,\Add+6(%edi)
1:  movb \Add+3(%edi),%al
    movb \Add+7(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 4(%esi),%al
    movb %al,\Add+3(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb (%esi),%dl
    movb %dl,\Add+7(%edi)
1:
.endif
.endm

.macro P16_Plot_8 Xflip,Add=0
.if \Xflip-1
    pushl %esi
    call PLOT8_8BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_8BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_8BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_8BplTile_X
    subl $8,%edi
.endif
.endm

.macro P8_Plot_4 Xflip,Add=0,Sprite=0
//  movb ColourBase,%dh
.if \Xflip-1
    movb \Add(%edi),%al
    movb \Add+4(%edi),%dl
    testb %al,%al
    jnz 1f
    movb (%esi),%al
.if \Sprite
    movb %cl,\Add(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 4(%esi),%dl
.if \Sprite
    movb %cl,\Add+4(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+4(%edi)
1:  movb \Add+1(%edi),%al
    movb \Add+5(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 1(%esi),%al
.if \Sprite
    movb %cl,\Add+1(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+1(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 5(%esi),%dl
.if \Sprite
    movb %cl,\Add+5(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+5(%edi)
1:  movb \Add+2(%edi),%al
    movb \Add+6(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 2(%esi),%al
.if \Sprite
    movb %cl,\Add+2(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+2(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 6(%esi),%dl
.if \Sprite
    movb %cl,\Add+6(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+6(%edi)
1:  movb \Add+3(%edi),%al
    movb \Add+7(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 3(%esi),%al
.if \Sprite
    movb %cl,\Add+3(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+3(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 7(%esi),%dl
.if \Sprite
    movb %cl,\Add+7(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+7(%edi)
1:
.else
    movb \Add(%edi),%al
    movb \Add+4(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 7(%esi),%al
.if \Sprite
    movb %cl,\Add(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 3(%esi),%dl
.if \Sprite
    movb %cl,\Add+4(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+4(%edi)
1:  movb \Add+1(%edi),%al
    movb \Add+5(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 6(%esi),%al
.if \Sprite
    movb %cl,\Add+1(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+1(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 2(%esi),%dl
.if \Sprite
    movb %cl,\Add+5(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+5(%edi)
1:  movb \Add+2(%edi),%al
    movb \Add+6(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 5(%esi),%al
.if \Sprite
    movb %cl,\Add+2(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+2(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb 1(%esi),%dl
.if \Sprite
    movb %cl,\Add+6(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+6(%edi)
1:  movb \Add+3(%edi),%al
    movb \Add+7(%edi),%dl
    testb %al,%al
    jnz 1f
    movb 4(%esi),%al
.if \Sprite
    movb %cl,\Add+3(%ebx)
.endif
    andb %dh,%al
    movb %al,\Add+3(%edi)
1:  testb %dl,%dl
    jnz 1f
    movb (%esi),%dl
.if \Sprite
    movb %cl,\Add+7(%ebx)
.endif
    andb %dh,%dl
    movb %dl,\Add+7(%edi)
1:
.endif
.endm

.macro P8_Plot_Half_4 Xflip,Add=0
//  movb ColourBase,%dh
.if \Xflip-1
    movb \Add(%edi),%al
    testb %al,%al
    jnz 1f
    movb (%esi),%al
    andb %dh,%al
    movb %al,\Add(%edi)
1:  movb \Add+1(%edi),%al
    testb %al,%al
    jnz 1f
    movb 2(%esi),%al
    andb %dh,%al
    movb %al,\Add+1(%edi)
1:  movb \Add+2(%edi),%al
    testb %al,%al
    jnz 1f
    movb 4(%esi),%al
    andb %dh,%al
    movb %al,\Add+2(%edi)
1:  movb \Add+3(%edi),%al
    testb %al,%al
    jnz 1f
    movb 6(%esi),%al
    andb %dh,%al
    movb %al,\Add+3(%edi)
1:
.else
    movb \Add(%edi),%al
    testb %al,%al
    jnz 1f
    movb 6(%esi),%al
    andb %dh,%al
    movb %al,\Add(%edi)
1:  movb \Add+1(%edi),%al
    testb %al,%al
    jnz 1f
    movb 4(%esi),%al
    andb %dh,%al
    movb %al,\Add+1(%edi)
1:  movb \Add+2(%edi),%al
    testb %al,%al
    jnz 1f
    movb 2(%esi),%al
    andb %dh,%al
    movb %al,\Add+2(%edi)
1:  movb \Add+3(%edi),%al
    testb %al,%al
    jnz 1f
    movb (%esi),%al
    andb %dh,%al
    movb %al,\Add+3(%edi)
1:
.endif
.endm

.macro P16_Plot_Half_4 Xflip,Add=0
.if \Xflip-1
    pushl %esi
    call PLOT8_4BplTile_Even
    popl %esi
    addl $64,%esi
    addl $4,%edi
    call PLOT8_4BplTile_Even
    subl $4,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_4BplTile_Even_X
    popl %esi
    addl $4,%edi
    call PLOT8_4BplTile_Even_X
    subl $4,%edi
.endif
.endm

.macro P16_Plot_4 Xflip,Add=0
.if \Xflip-1
    pushl %esi
    call PLOT8_4BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_4BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_4BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_4BplTile_X
    subl $8,%edi
.endif
.endm

.macro P8_Plot_2 Xflip,Add=0
    P8_Plot_4 Xflip=\Xflip,Add=\Add
.endm

.macro P8_Plot_Half_2 Xflip,Add=0
    P8_Plot_Half_4 Xflip=\Xflip,Add=\Add
.endm

.macro P16_Plot_Half_2 Xflip,Add=0
.if \Xflip-1
    pushl %esi
    call PLOT8_2BplTile_Even
    popl %esi
    addl $64,%esi
    addl $4,%edi
    call PLOT8_2BplTile_Even
    subl $4,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_2BplTile_Even_X
    popl %esi
    addl $4,%edi
    call PLOT8_2BplTile_Even_X
    subl $4,%edi
.endif
.endm

.macro P16_Plot_2 Xflip,Add=0
.if \Xflip-1
    pushl %esi
    call PLOT8_2BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_2BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_2BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_2BplTile_X
    subl $8,%edi
.endif
.endm

ALIGNC
PLOT8_8BplTile:
    P8_Plot_8 Xflip=0
    ret

ALIGNC
PLOT8_8BplTile_X:
    P8_Plot_8 Xflip=1
    ret

ALIGNC
PLOT8_4BplTile:
    P8_Plot_4 Xflip=0
    ret

ALIGNC
PLOT8_4BplTile_X:
    P8_Plot_4 Xflip=1
    ret

ALIGNC
PLOT8_2BplTile:
    P8_Plot_2 Xflip=0
    ret

ALIGNC
PLOT8_2BplTile_X:
    P8_Plot_2 Xflip=1
    ret

PLOT8_4BplTile_Even:
    P8_Plot_Half_4 Xflip=0
    ret

PLOT8_4BplTile_Even_X:
    P8_Plot_Half_4 Xflip=1
    ret

PLOT8_2BplTile_Even:
    P8_Plot_Half_2 Xflip=0
    ret

PLOT8_2BplTile_Even_X:
    P8_Plot_Half_2 Xflip=1
    ret

ALIGNC
PLOT16_4BplTile_Even:
    P16_Plot_Half_4 Xflip=0
    ret

ALIGNC
PLOT16_4BplTile_Even_X:
    P16_Plot_Half_4 Xflip=1
    ret

ALIGNC
PLOT16_2BplTile_Even:
    P16_Plot_Half_2 Xflip=0
    ret

ALIGNC
PLOT16_2BplTile_Even_X:
    P16_Plot_Half_2 Xflip=1
    ret

ALIGNC
PLOT16_8BplTile:
    P16_Plot_8 Xflip=0
    ret

ALIGNC
PLOT16_8BplTile_X:
    P16_Plot_8 Xflip=1
    ret

ALIGNC
PLOT16_4BplTile:
    P16_Plot_4 Xflip=0
    ret

ALIGNC
PLOT16_4BplTile_X:
    P16_Plot_4 Xflip=1
    ret

ALIGNC
PLOT16_2BplTile:
    P16_Plot_2 Xflip=0
    ret

ALIGNC
PLOT16_2BplTile_X:
    P16_Plot_2 Xflip=1
    ret

ALIGNC
PLOT8_4BplTile_Sprite:
    addl _TileCache4,%esi
    P8_Plot_4 Xflip=0 Sprite=1
    ret

ALIGNC
PLOT8_4BplTile_Sprite_X:
    addl _TileCache4,%esi
    P8_Plot_4 Xflip=1 Sprite=1
    ret

ALIGNC
PLOT16_4BplSprite:
//  movb ColourBase,%dh
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite
ALIGNC
PLOT16_4BplSprite_X:
//  movb ColourBase,%dh
    pushl %esi
    addl $64,%esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite_X

ALIGNC
PLOT32_4BplSprite:
//  movb ColourBase,%dh
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite
ALIGNC
PLOT32_4BplSprite_X:
//  movb ColourBase,%dh
    pushl %esi
    addl $64*3,%esi
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite_X

ALIGNC
PLOT64_4BplSprite:
//  movb ColourBase,%dh
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite
    popl %esi
    addl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite
ALIGNC
PLOT64_4BplSprite_X:
//  movb ColourBase,%dh
    pushl %esi
    addl $64*7,%esi
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    pushl %esi
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    addl $8,%ebx
    call PLOT8_4BplTile_Sprite_X
    popl %esi
    addl $8,%edi
    addl $8,%ebx
    jmp PLOT8_4BplTile_Sprite_X

ALIGNC
.globl Invalidate_Tile_Caches
Invalidate_Tile_Caches:
    movl $-2,Tile_Recache_Set_Begin
    movl $-2,Tile_Recache_Set_End
    ret

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
